#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Tests for zero-order anaerobic MBR-MEC model
"""
import pytest

from pyomo.environ import (
    ConcreteModel,
    Constraint,
    value,
    Var,
    assert_optimal_termination,
    units as pyunits,
    Block,
    Param,
)
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock
from idaes.core.solvers import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.testing import initialization_tester
from idaes.core import UnitModelCostingBlock

from watertap.unit_models.zero_order import AnaerobicMBRMECZO
from watertap.core.wt_database import Database
from watertap.core.zero_order_properties import WaterParameterBlock
from watertap.costing.zero_order_costing import ZeroOrderCosting

solver = get_solver()


class TestAnaerobicMBRMECZO:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.db = Database()

        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.params = WaterParameterBlock(
            solute_list=[
                "cod",
                "nonbiodegradable_cod",
                "ammonium_as_nitrogen",
                "phosphate_as_phosphorous",
            ]
        )

        m.fs.unit = AnaerobicMBRMECZO(property_package=m.fs.params, database=m.db)

        m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(0.043642594)
        m.fs.unit.inlet.flow_mass_comp[0, "cod"].fix(1.00625e-4)
        m.fs.unit.inlet.flow_mass_comp[0, "nonbiodegradable_cod"].fix(1e-20)
        m.fs.unit.inlet.flow_mass_comp[0, "ammonium_as_nitrogen"].fix(4.59375e-06)
        m.fs.unit.inlet.flow_mass_comp[0, "phosphate_as_phosphorous"].fix(2.1875e-06)

        return m

    @pytest.mark.unit
    def test_build(self, model):
        assert model.fs.unit.config.database == model.db

        assert isinstance(model.fs.unit.lift_height, Param)
        assert isinstance(model.fs.unit.eta_pump, Param)
        assert isinstance(model.fs.unit.eta_motor, Param)
        assert isinstance(model.fs.unit.electricity, Var)
        assert isinstance(model.fs.unit.electricity_consumption, Constraint)

    @pytest.mark.component
    def test_load_parameters(self, model):
        data = model.db.get_unit_operation_parameters("anaerobic_mbr_mec")

        model.fs.unit.load_parameters_from_database(use_default_removal=True)

        assert model.fs.unit.recovery_frac_mass_H2O[0].fixed
        assert (
            model.fs.unit.recovery_frac_mass_H2O[0].value
            == data["recovery_frac_mass_H2O"]["value"]
        )

        for (t, j), v in model.fs.unit.removal_frac_mass_comp.items():
            assert v.fixed
            if j not in data["removal_frac_mass_comp"].keys():
                assert v.value == data["default_removal_frac_mass_comp"]["value"]
            else:
                assert v.value == data["removal_frac_mass_comp"][j]["value"]

    @pytest.mark.component
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model.fs.unit) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model.fs.unit)

    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        assert pytest.approx(3780 / 3600 / 24 / 1000, rel=1e-5) == value(
            model.fs.unit.properties_in[0].flow_vol
        )
        assert pytest.approx(2.3, rel=1e-5) == value(
            model.fs.unit.properties_in[0].conc_mass_comp["cod"]
        )
        assert pytest.approx(0, rel=1e-5) == value(
            model.fs.unit.properties_in[0].conc_mass_comp["nonbiodegradable_cod"]
        )
        assert pytest.approx(0.105, rel=1e-5) == value(
            model.fs.unit.properties_in[0].conc_mass_comp["ammonium_as_nitrogen"]
        )
        assert pytest.approx(0.05, rel=1e-5) == value(
            model.fs.unit.properties_in[0].conc_mass_comp["phosphate_as_phosphorous"]
        )

        assert pytest.approx(1500 / 3600 / 24 / 1000, rel=1e-2) == value(
            model.fs.unit.properties_treated[0].flow_vol
        )
        assert pytest.approx(2.895771, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["cod"]
        )
        assert pytest.approx(1.5972836e-09, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["nonbiodegradable_cod"]
        )

        assert pytest.approx(2280 / 3600 / 24 / 1000, rel=1e-2) == value(
            model.fs.unit.properties_byproduct[0].flow_vol
        )
        assert pytest.approx(1.0521862e-09, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].conc_mass_comp["cod"]
        )
        assert pytest.approx(1.907545, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].conc_mass_comp["nonbiodegradable_cod"]
        )
        assert pytest.approx(4.347, rel=1e-3) == value(
            pyunits.convert(
                model.fs.unit.properties_byproduct[0].flow_mass_comp[
                    "nonbiodegradable_cod"
                ],
                to_units=pyunits.kg / pyunits.day,
            )
        )
        assert pytest.approx(0.0161213, rel=1e-5) == value(model.fs.unit.electricity[0])

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, model):
        for j in model.fs.params.component_list:
            assert 1e-6 >= abs(
                value(
                    model.fs.unit.inlet.flow_mass_comp[0, j]
                    + sum(
                        model.fs.unit.generation_rxn_comp[0, r, j]
                        for r in model.fs.unit.reaction_set
                    )
                    - model.fs.unit.treated.flow_mass_comp[0, j]
                    - model.fs.unit.byproduct.flow_mass_comp[0, j]
                )
            )

    @pytest.mark.component
    def test_report(self, model):
        model.fs.unit.report()


db = Database()
params = db._get_technology("anaerobic_mbr_mec")


class Test_AnMBRMEC_ZO_subtype:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()

        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.params = WaterParameterBlock(solute_list=["cod", "nonbiodegradable_cod"])

        m.fs.unit = AnaerobicMBRMECZO(property_package=m.fs.params, database=db)

        return m

    @pytest.mark.parametrize("subtype", [params.keys()])
    @pytest.mark.component
    def test_load_parameters(self, model, subtype):
        model.fs.unit.config.process_subtype = subtype
        data = db.get_unit_operation_parameters("anaerobic_mbr_mec", subtype=subtype)

        model.fs.unit.load_parameters_from_database()

        for (t, j), v in model.fs.unit.removal_frac_mass_comp.items():
            assert v.fixed
            assert v.value == data["removal_frac_mass_comp"][j]["value"]


@pytest.mark.unit
def test_ffCOD_not_in_solute_list():
    model = ConcreteModel()
    model.db = Database()

    model.fs = FlowsheetBlock(dynamic=False)
    model.fs.params = WaterParameterBlock(solute_list=["cod"])
    with pytest.raises(
        ValueError,
        match="nonbiodegradable_cod must be included in the solute list since"
        " this unit model converts cod to nonbiodegradable_cod.",
    ):
        model.fs.unit = AnaerobicMBRMECZO(
            property_package=model.fs.params, database=model.db
        )


@pytest.mark.unit
def test_COD_not_in_solute_list():
    model = ConcreteModel()
    model.db = Database()

    model.fs = FlowsheetBlock(dynamic=False)
    model.fs.params = WaterParameterBlock(solute_list=["nonbiodegradable_cod"])
    with pytest.raises(
        ValueError,
        match="fs.unit - key_reactant cod for reaction cod_to_nonbiodegradable_cod "
        "is not in the component list used by the assigned property package.",
    ):
        model.fs.unit = AnaerobicMBRMECZO(
            property_package=model.fs.params, database=model.db
        )


def test_costing():
    m = ConcreteModel()
    m.db = Database()

    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.params = WaterParameterBlock(
        solute_list=[
            "cod",
            "nonbiodegradable_cod",
            "ammonium_as_nitrogen",
            "phosphate_as_phosphorous",
        ]
    )

    m.fs.costing = ZeroOrderCosting()

    m.fs.unit1 = AnaerobicMBRMECZO(property_package=m.fs.params, database=m.db)

    m.fs.unit1.inlet.flow_mass_comp[0, "H2O"].fix(0.043642594)
    m.fs.unit1.inlet.flow_mass_comp[0, "cod"].fix(1.00625e-4)
    m.fs.unit1.inlet.flow_mass_comp[0, "nonbiodegradable_cod"].fix(1e-20)
    m.fs.unit1.inlet.flow_mass_comp[0, "ammonium_as_nitrogen"].fix(4.59375e-06)
    m.fs.unit1.inlet.flow_mass_comp[0, "phosphate_as_phosphorous"].fix(2.1875e-06)

    m.fs.unit1.load_parameters_from_database(use_default_removal=True)
    assert degrees_of_freedom(m.fs.unit1) == 0

    m.fs.unit1.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)

    assert isinstance(m.fs.costing.anaerobic_mbr_mec, Block)
    assert isinstance(m.fs.costing.anaerobic_mbr_mec.unit_capex, Var)
    assert isinstance(m.fs.costing.anaerobic_mbr_mec.unit_opex, Var)
    assert isinstance(m.fs.unit1.costing.capital_cost, Var)
    assert isinstance(m.fs.unit1.costing.capital_cost_constraint, Constraint)

    assert_units_consistent(m.fs)
    assert degrees_of_freedom(m.fs.unit1) == 0

    assert m.fs.unit1.electricity[0] in m.fs.costing._registered_flows["electricity"]



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

from .watertap_costing_package import WaterTAPCosting
from .zero_order_costing import ZeroOrderCosting
from .multiple_choice_costing_block import MultiUnitModelCostingBlock

from .util import (
    register_costing_parameter_block,
    make_capital_cost_var,
    make_fixed_operating_cost_var,
    cost_by_flow_volume,
    cost_membrane,
    cost_rectifier,
)

from .unit_models.crystallizer import CrystallizerCostType
from .unit_models.energy_recovery_device import EnergyRecoveryDeviceType
from .unit_models.mixer import MixerType
from .unit_models.pump import PumpType
from .unit_models.reverse_osmosis import ROType



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

from pyomo.common.config import ConfigValue, In
from pyomo.environ import Set
from idaes.core import (
    declare_process_block_class,
    DistributedVars,
    FlowDirection,
    useDefault,
)
from idaes.core.base.control_volume1d import ControlVolume1DBlockData
from idaes.core.util import scaling as iscale
from idaes.core.util.misc import add_object_reference
import idaes.logger as idaeslog

from watertap.core.membrane_channel_base import (
    MembraneChannelMixin,
    PressureChangeType,
    CONFIG_Template as Base_CONFIG_Template,
)

CONFIG_Template = Base_CONFIG_Template()

CONFIG_Template.declare(
    "area_definition",
    ConfigValue(
        default=DistributedVars.uniform,
        domain=In(DistributedVars),
        description="Argument for defining form of area variable",
        doc="""Argument defining whether area variable should be spatially
variant or not. **default** - DistributedVars.uniform.
**Valid values:** {
DistributedVars.uniform - area does not vary across spatial domain,
DistributedVars.variant - area can vary over the domain and is indexed
by time and space.}""",
    ),
)

CONFIG_Template.declare(
    "transformation_method",
    ConfigValue(
        default=useDefault,
        description="Discretization method to use for DAE transformation",
        doc="""Discretization method to use for DAE transformation. See Pyomo
documentation for supported transformations.""",
    ),
)

CONFIG_Template.declare(
    "transformation_scheme",
    ConfigValue(
        default=useDefault,
        description="Discretization scheme to use for DAE transformation",
        doc="""Discretization scheme to use when transforming domain. See
Pyomo documentation for supported schemes.""",
    ),
)

CONFIG_Template.declare(
    "finite_elements",
    ConfigValue(
        default=10,
        domain=int,
        description="Number of finite elements in length domain",
        doc="""Number of finite elements to use when discretizing length 
        domain (default=10)""",
    ),
)

CONFIG_Template.declare(
    "collocation_points",
    ConfigValue(
        default=5,
        domain=int,
        description="Number of collocation points per finite element",
        doc="""Number of collocation points to use per finite element when
        discretizing length domain (default=5)""",
    ),
)


@declare_process_block_class("MembraneChannel1DBlock")
class MembraneChannel1DBlockData(MembraneChannelMixin, ControlVolume1DBlockData):
    def _skip_element(self, x):
        if self.config.transformation_scheme != "FORWARD":
            return x == self.length_domain.first()
        else:
            return x == self.length_domain.last()

    def apply_transformation(self, *args, **kwargs):
        super().apply_transformation(*args, **kwargs)
        self.difference_elements = Set(
            ordered=True,
            initialize=(x for x in self.length_domain if not self._skip_element(x)),
        )
        self._set_nfe()

    def add_geometry(
        self, length_var, width_var, flow_direction=FlowDirection.forward, **kwargs
    ):
        """
        Method to create spatial domain and volume Var in ControlVolume.

        Args:
            length_var - An external variable to use for the length of
                         the channel. If a variable is provided, a
                         reference will be made to this in place of the length
                         Var.

            width_var - An external variable to use for the width of
                         the channel. If a variable is provided, a
                         reference will be made to this in place of the length
                         Var.
            flow_direction - argument indicating direction of material flow
                            relative to length domain. Valid values:
                                - FlowDirection.forward (default), flow goes
                                  from 0 to 1.
                                - FlowDirection.backward, flow goes from 1 to 0
            length_domain - (optional) a ContinuousSet to use as the length
                            domain for the ControlVolume. If not provided, a
                            new ContinuousSet will be created (default=None).
                            ContinuousSet should be normalized to run between
                            0 and 1.
            length_domain_set - (optional) list of point to use to initialize
                            a new ContinuousSet if length_domain is not
                            provided (default = [0.0, 1.0]).

        Returns:
            None
        """
        super().add_geometry(
            length_var=length_var, flow_direction=flow_direction, **kwargs
        )
        self._add_var_reference(width_var, "width", "width_var")

    def add_state_blocks(self, has_phase_equilibrium=None):
        """
        This method constructs the state blocks for the
        control volume.

        Args:
            has_phase_equilibrium: indicates whether equilibrium calculations
                                    will be required in state blocks
        Returns:
            None
        """
        super().add_state_blocks(has_phase_equilibrium=has_phase_equilibrium)
        self._add_interface_stateblock(has_phase_equilibrium)

    def _add_pressure_change(self, pressure_change_type=PressureChangeType.calculated):
        add_object_reference(self, "dP_dx", self.deltaP)

    def initialize(
        self,
        state_args=None,
        outlvl=idaeslog.NOTSET,
        optarg=None,
        solver=None,
        hold_state=True,
        initialize_guess=None,
    ):
        """
        Initialization routine for the membrane channel control volume

        Keyword Arguments:
            state_args : a dict of arguments to be passed to the property
                         package(s) to provide an initial state for
                         initialization (see documentation of the specific
                         property package) (default = {}).
            outlvl : sets output log level of initialization routine
            optarg : solver options dictionary object (default=None, use
                     default solver options)
            solver : str indicating which solver to use during
                     initialization (default = None)
            hold_state : flag indicating whether the initialization routine
                     should unfix any state variables fixed during
                     initialization, **default** - True. **Valid values:**
                     **True** - states variables are not unfixed, and a dict of
                     returned containing flags for which states were fixed
                     during initialization, **False** - state variables are
                     unfixed after initialization by calling the release_state
                     method.
            initialize_guess : a dict of guesses for solvent_recovery, solute_recovery,
                     and cp_modulus. These guesses offset the initial values
                     for the retentate, permeate, and membrane interface
                     state blocks from the inlet feed
                     (default =
                     {'deltaP': -1e4,
                     'solvent_recovery': 0.5,
                     'solute_recovery': 0.01,
                     'cp_modulus': 1.1})

        Returns:
            If hold_states is True, returns a dict containing flags for which
            states were fixed during initialization.
        """

        # Get inlet state if not provided
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="control_volume")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="control_volume")

        state_args = self._get_state_args(initialize_guess, state_args)

        # intialize self.properties
        state_args_properties_in = state_args["feed_side"]
        if self._flow_direction == FlowDirection.forward:
            state_args_properties_out = state_args["retentate"]
        else:
            state_args_properties_out = state_args["permeate"]

        source_flags = super().initialize(
            state_args=state_args_properties_in,
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            hold_state=True,
        )

        state_args_interface = self._get_state_args_interface(
            initialize_guess, state_args_properties_in, state_args_properties_out
        )
        self.properties_interface.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=state_args_interface,
        )

        if hold_state:
            return source_flags
        else:
            self.release_state(source_flags, outlvl)

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()

        # setting scaling factors for variables

        # will not override if the user provides the scaling factor
        ## default of 1 set by ControlVolume1D
        if iscale.get_scaling_factor(self.area) == 1:
            iscale.set_scaling_factor(self.area, 100)

        if hasattr(self, "pressure_change_total"):
            for v in self.pressure_change_total.values():
                if iscale.get_scaling_factor(v) is None:
                    iscale.set_scaling_factor(v, 1e-4)

        if hasattr(self, "dP_dx"):
            for v in self.pressure_dx.values():
                iscale.set_scaling_factor(v, 1e-5)



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import pytest
from pyomo.environ import (
    ConcreteModel,
    value,
    Var,
    Constraint,
    assert_optimal_termination,
)
from pyomo.network import Port
from idaes.core import (
    FlowsheetBlock,
    MaterialBalanceType,
    EnergyBalanceType,
    MomentumBalanceType,
)
from watertap.unit_models.nanofiltration_ZO import NanofiltrationZO
from idaes.models.properties.modular_properties.base.generic_property import (
    GenericParameterBlock,
)
from watertap.property_models.seawater_ion_generic import configuration
import watertap.property_models.seawater_ion_prop_pack as props

from watertap.core.util.initialization import assert_no_degrees_of_freedom
from pyomo.util.check_units import assert_units_consistent

from idaes.core.solvers import get_solver
from idaes.core.util.model_statistics import (
    number_variables,
    number_total_constraints,
    number_unused_variables,
)
from idaes.core.util.testing import initialization_tester
from idaes.core.util.scaling import (
    calculate_scaling_factors,
    constraint_scaling_transform,
    unscaled_variables_generator,
    unscaled_constraints_generator,
    badly_scaled_var_generator,
)

# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()

# -----------------------------------------------------------------------------
@pytest.mark.unit
def test_config():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = props.PropParameterBlock()
    m.fs.unit = NanofiltrationZO(property_package=m.fs.properties)

    assert len(m.fs.unit.config) == 9

    assert not m.fs.unit.config.dynamic
    assert not m.fs.unit.config.has_holdup
    assert m.fs.unit.config.material_balance_type == MaterialBalanceType.useDefault
    assert m.fs.unit.config.energy_balance_type == EnergyBalanceType.none
    assert m.fs.unit.config.momentum_balance_type == MomentumBalanceType.pressureTotal
    assert not m.fs.unit.config.has_pressure_change
    assert m.fs.unit.config.property_package is m.fs.properties


@pytest.mark.unit
def test_option_has_pressure_change():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = props.PropParameterBlock()
    m.fs.unit = NanofiltrationZO(
        property_package=m.fs.properties, has_pressure_change=True
    )

    assert isinstance(m.fs.unit.feed_side.deltaP, Var)
    assert isinstance(m.fs.unit.deltaP, Var)


class TestNanofiltration:
    @pytest.fixture(scope="class")
    def unit_frame(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.properties = props.PropParameterBlock()

        m.fs.unit = NanofiltrationZO(property_package=m.fs.properties)

        # fully specify system
        feed_flow_mass = 1
        feed_mass_frac = {
            "Na": 11122e-6,
            "Ca": 382e-6,
            "Mg": 1394e-6,
            "SO4": 2136e-6,
            "Cl": 20316.88e-6,
        }
        m.fs.unit.feed_side.properties_in[0].flow_mass_phase_comp["Liq", "H2O"].fix(
            feed_flow_mass * (1 - sum(x for x in feed_mass_frac.values()))
        )
        for j in feed_mass_frac:
            m.fs.unit.feed_side.properties_in[0].flow_mass_phase_comp["Liq", j].fix(
                feed_flow_mass * feed_mass_frac[j]
            )
        m.fs.unit.feed_side.properties_in[0].pressure.fix(101325)
        m.fs.unit.feed_side.properties_in[0].temperature.fix(298.15)

        m.fs.unit.flux_vol_solvent.fix(1.67e-6)
        m.fs.unit.area.fix(500)
        m.fs.unit.properties_permeate[0].pressure.fix(101325)

        m.fs.unit.rejection_phase_comp[0, "Liq", "Na"].fix(0.01)
        m.fs.unit.rejection_phase_comp[0, "Liq", "Ca"].fix(0.79)
        m.fs.unit.rejection_phase_comp[0, "Liq", "Mg"].fix(0.94)
        m.fs.unit.rejection_phase_comp[0, "Liq", "SO4"].fix(0.87)
        m.fs.unit.rejection_phase_comp[
            0, "Liq", "Cl"
        ] = 0.15  # guess, but electroneutrality enforced below
        charge_comp = {"Na": 1, "Ca": 2, "Mg": 2, "SO4": -2, "Cl": -1}
        m.fs.unit.eq_electroneutrality = Constraint(
            expr=0
            == sum(
                charge_comp[j]
                * m.fs.unit.feed_side.properties_out[0].conc_mol_phase_comp["Liq", j]
                for j in charge_comp
            )
        )
        constraint_scaling_transform(m.fs.unit.eq_electroneutrality, 1)

        return m

    @pytest.mark.unit
    def test_build(self, unit_frame):
        m = unit_frame

        # test ports
        port_lst = ["inlet", "retentate", "permeate"]
        for port_str in port_lst:
            port = getattr(m.fs.unit, port_str)
            assert (
                len(port.vars) == 3
            )  # number of state variables for NaCl property package
            assert isinstance(port, Port)

        # test statistics
        assert number_variables(m) == 89
        assert number_total_constraints(m) == 74
        assert number_unused_variables(m) == 1

    @pytest.mark.unit
    def test_dof(self, unit_frame):
        m = unit_frame
        assert_no_degrees_of_freedom(m)

    @pytest.mark.unit
    def test_calculate_scaling(self, unit_frame):
        m = unit_frame

        m.fs.properties.set_default_scaling(
            "flow_mass_phase_comp", 1, index=("Liq", "H2O")
        )
        m.fs.properties.set_default_scaling(
            "flow_mass_phase_comp", 1e2, index=("Liq", "Na")
        )
        m.fs.properties.set_default_scaling(
            "flow_mass_phase_comp", 1e4, index=("Liq", "Ca")
        )
        m.fs.properties.set_default_scaling(
            "flow_mass_phase_comp", 1e3, index=("Liq", "Mg")
        )
        m.fs.properties.set_default_scaling(
            "flow_mass_phase_comp", 1e3, index=("Liq", "SO4")
        )
        m.fs.properties.set_default_scaling(
            "flow_mass_phase_comp", 1e2, index=("Liq", "Cl")
        )
        calculate_scaling_factors(m)

        # check that all variables have scaling factors
        unscaled_var_list = list(unscaled_variables_generator(m))
        assert len(unscaled_var_list) == 0
        # check that all constraints have been scaled
        unscaled_constraint_list = list(unscaled_constraints_generator(m))
        assert len(unscaled_constraint_list) == 0

    @pytest.mark.component
    def test_initialize(self, unit_frame):
        initialization_tester(unit_frame)

    @pytest.mark.component
    def test_var_scaling(self, unit_frame):
        m = unit_frame
        badly_scaled_var_lst = list(badly_scaled_var_generator(m))
        assert badly_scaled_var_lst == []

    @pytest.mark.component
    def test_solve(self, unit_frame):
        m = unit_frame
        solver.options = {"nlp_scaling_method": "user-scaling"}
        results = solver.solve(m)

        # Check for optimal solution
        assert_optimal_termination(results)

    @pytest.mark.component
    def test_conservation(self, unit_frame):
        m = unit_frame
        b = m.fs.unit
        comp_lst = m.fs.properties.component_list

        flow_mass_inlet = sum(
            b.feed_side.properties_in[0].flow_mass_phase_comp["Liq", j]
            for j in comp_lst
        )
        flow_mass_retentate = sum(
            b.feed_side.properties_out[0].flow_mass_phase_comp["Liq", j]
            for j in comp_lst
        )
        flow_mass_permeate = sum(
            b.properties_permeate[0].flow_mass_phase_comp["Liq", j] for j in comp_lst
        )

        assert (
            abs(value(flow_mass_inlet - flow_mass_retentate - flow_mass_permeate))
            <= 1e-5
        )

    @pytest.mark.component
    def test_solution(self, unit_frame):
        m = unit_frame
        assert pytest.approx(0.8350, rel=1e-3) == value(
            m.fs.unit.properties_permeate[0].flow_mass_phase_comp["Liq", "H2O"]
        )
        assert pytest.approx(2.002, rel=1e-3) == value(
            m.fs.unit.properties_permeate[0].conc_mol_phase_comp["Liq", "Ca"]
        )
        assert pytest.approx(484.0, rel=1e-3) == value(
            m.fs.unit.properties_permeate[0].conc_mol_phase_comp["Liq", "Cl"]
        )
        assert pytest.approx(3.441, rel=1e-3) == value(
            m.fs.unit.properties_permeate[0].conc_mol_phase_comp["Liq", "Mg"]
        )
        assert pytest.approx(478.9, rel=1e-3) == value(
            m.fs.unit.properties_permeate[0].conc_mol_phase_comp["Liq", "Na"]
        )
        assert pytest.approx(2.891, rel=1e-3) == value(
            m.fs.unit.properties_permeate[0].conc_mol_phase_comp["Liq", "SO4"]
        )
        assert pytest.approx(0.86, rel=1e-3) == value(
            m.fs.unit.recovery_vol_phase[0, "Liq"]
        )

    @pytest.mark.unit
    def test_report(self, unit_frame):
        unit_frame.fs.unit.report()

    @pytest.mark.component
    def test_NF_with_generic_property_model(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.properties = GenericParameterBlock(**configuration)
        m.fs.unit = NanofiltrationZO(
            property_package=m.fs.properties, has_pressure_change=False
        )

        # fully specify system
        m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "Na_+"].fix(0.008845)
        m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "Ca_2+"].fix(0.000174)
        m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "Mg_2+"].fix(0.001049)
        m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "SO4_2-"].fix(0.000407)
        m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "Cl_-"].fix(0.010479)
        m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "H2O"].fix(0.979046)
        m.fs.unit.feed_side.properties_in[0].pressure.fix(4e5)
        m.fs.unit.feed_side.properties_in[0].temperature.fix(298.15)

        m.fs.unit.flux_vol_solvent.fix(1.67e-6)
        m.fs.unit.recovery_vol_phase.fix(0.86)
        m.fs.unit.properties_permeate[0].pressure.fix(101325)

        m.fs.unit.rejection_phase_comp[0, "Liq", "Na_+"].fix(0.01)
        m.fs.unit.rejection_phase_comp[0, "Liq", "Ca_2+"].fix(0.79)
        m.fs.unit.rejection_phase_comp[0, "Liq", "Mg_2+"].fix(0.94)
        m.fs.unit.rejection_phase_comp[0, "Liq", "SO4_2-"].fix(0.87)
        m.fs.unit.rejection_phase_comp[
            0, "Liq", "Cl_-"
        ] = 0.15  # guess, but electroneutrality enforced below
        charge_comp = {"Na_+": 1, "Ca_2+": 2, "Mg_2+": 2, "SO4_2-": -2, "Cl_-": -1}
        m.fs.unit.eq_electroneutrality = Constraint(
            expr=0
            == sum(
                charge_comp[j]
                * m.fs.unit.feed_side.properties_out[0].conc_mol_phase_comp["Liq", j]
                for j in charge_comp
            )
        )
        constraint_scaling_transform(m.fs.unit.eq_electroneutrality, 1)

        assert_units_consistent(m)

        assert_no_degrees_of_freedom(m)

        calculate_scaling_factors(m)

        initialization_tester(m)

        results = solver.solve(m)

        # Check for optimal solution
        assert_optimal_termination(results)

        assert pytest.approx(0.868, rel=1e-3) == value(
            m.fs.unit.properties_permeate[0].flow_mass_phase_comp["Liq", "H2O"]
            / m.fs.unit.feed_side.properties_in[0].flow_mass_phase_comp["Liq", "H2O"]
        )
        assert pytest.approx(1.978, rel=1e-3) == value(
            m.fs.unit.properties_permeate[0].conc_mol_phase_comp["Liq", "Ca_2+"]
        )
        assert pytest.approx(479.1, rel=1e-3) == value(
            m.fs.unit.properties_permeate[0].conc_mol_phase_comp["Liq", "Cl_-"]
        )
        assert pytest.approx(3.407, rel=1e-3) == value(
            m.fs.unit.properties_permeate[0].conc_mol_phase_comp["Liq", "Mg_2+"]
        )
        assert pytest.approx(473.9, rel=1e-3) == value(
            m.fs.unit.properties_permeate[0].conc_mol_phase_comp["Liq", "Na_+"]
        )
        assert pytest.approx(2.864, rel=1e-3) == value(
            m.fs.unit.properties_permeate[0].conc_mol_phase_comp["Liq", "SO4_2-"]
        )



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
from watertap.tools.parameter_sweep import (
    LinearSample,
    parameter_sweep,
)
import watertap.examples.flowsheets.case_studies.electroNP.electroNP_flowsheet as electroNP_flowsheet
from pyomo.environ import units as pyunits


def set_up_sensitivity(m):
    outputs = {}
    optimize_kwargs = {"fail_flag": False}
    opt_function = electroNP_flowsheet.solve

    # create outputs
    outputs["LCOW"] = m.fs.costing.LCOW
    outputs["ElectroNP Capital Cost"] = m.fs.electroNP.costing.capital_cost
    outputs["Electricity"] = pyunits.convert(
        m.fs.costing.aggregate_flow_costs["electricity"] / m.fs.AD.inlet.flow_vol[0],
        to_units=pyunits.USD_2018 / pyunits.m**3,
    )
    outputs[
        "S_PO4 Concentration, treated effluent stream"
    ] = m.fs.electroNP.treated.conc_mass_comp[0, "S_PO4"]

    return outputs, optimize_kwargs, opt_function


def run_analysis(case_num=1, nx=11, interpolate_nan_outputs=True, output_filename=None):

    if output_filename is None:
        output_filename = "sensitivity_" + str(case_num) + ".csv"

    m = electroNP_flowsheet.build_flowsheet()[0]

    outputs, optimize_kwargs, opt_function = set_up_sensitivity(m)

    sweep_params = {}
    if case_num == 1:
        # sensitivity analysis
        m.fs.costing.electroNP.sizing_cost.unfix()
        sweep_params["sizing_cost"] = LinearSample(
            m.fs.costing.electroNP.sizing_cost, 1, 30, nx
        )
    elif case_num == 2:
        m.fs.costing.electroNP.sizing_cost.unfix()
        sweep_params["sizing_cost"] = LinearSample(
            m.fs.costing.electroNP.sizing_cost, 0, 30, nx
        )
        m.fs.costing.electroNP.HRT.unfix()
        sweep_params["HRT"] = LinearSample(m.fs.costing.electroNP.HRT, 0.1, 5, nx)
    elif case_num == 3:
        sweep_params["energy_consumption"] = LinearSample(
            m.fs.electroNP.energy_electric_flow_mass, 0.04, 2.5, nx
        )
    elif case_num == 4:
        sweep_params["energy_consumption"] = LinearSample(
            m.fs.electroNP.energy_electric_flow_mass, 0, 2.5, nx
        )
        sweep_params["electricity_cost"] = LinearSample(
            m.fs.costing.electricity_cost, 0.02, 0.3, nx
        )
    elif case_num == 5:
        sweep_params["electricity_cost"] = LinearSample(
            m.fs.costing.electricity_cost, 0.02, 0.3, nx
        )
        sweep_params["MgCl2_cost"] = LinearSample(
            m.fs.costing.magnesium_chloride_cost, 0.06, 0.1, nx
        )
    elif case_num == 6:
        sweep_params["energy_consumption"] = LinearSample(
            m.fs.electroNP.energy_electric_flow_mass, 0, 2.5, nx
        )
        sweep_params["MgCl2_dosage"] = LinearSample(
            m.fs.electroNP.magnesium_chloride_dosage, 0.1, 0.5, nx
        )
    elif case_num == 7:
        sweep_params["S_IP_concentration"] = LinearSample(
            m.fs.AD.inlet.conc_mass_comp[0, "S_IP"], 0.01, 0.5, nx
        )
    elif case_num == 8:
        sweep_params["S_IP_concentration"] = LinearSample(
            m.fs.AD.inlet.conc_mass_comp[0, "S_IP"], 0.01, 0.5, nx
        )
        sweep_params["X_PHA_concentration"] = LinearSample(
            m.fs.AD.inlet.conc_mass_comp[0, "X_PHA"], 0, 0.5, nx
        )
    elif case_num == 9:
        sweep_params["X_PAO_concentration"] = LinearSample(
            m.fs.AD.inlet.conc_mass_comp[0, "X_PAO"], 3.2, 3.8, nx
        )
        sweep_params["X_PHA_concentration"] = LinearSample(
            m.fs.AD.inlet.conc_mass_comp[0, "X_PHA"], 1e-6, 0.5, nx
        )
    elif case_num == 10:
        sweep_params["X_PAO_concentration"] = LinearSample(
            m.fs.AD.inlet.conc_mass_comp[0, "X_PHA"], 0, 0.5, nx
        )
        sweep_params["X_PP_concentration"] = LinearSample(
            m.fs.AD.inlet.conc_mass_comp[0, "X_PP"], 0.8, 1.2, nx
        )
    else:
        raise ValueError(f"{case_num} is not yet implemented")

    global_results = parameter_sweep(
        m,
        sweep_params,
        outputs,
        csv_results_file_name=output_filename,
        optimize_function=opt_function,
        optimize_kwargs=optimize_kwargs,
        interpolate_nan_outputs=interpolate_nan_outputs,
    )

    return global_results, sweep_params, m


if __name__ == "__main__":
    results, sweep_params, m = run_analysis()



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import pyomo.environ as pyo
from ..util import (
    register_costing_parameter_block,
    make_capital_cost_var,
    make_fixed_operating_cost_var,
)


def build_hcl_cost_param_block(blk):

    blk.cost = pyo.Param(
        mutable=True,
        initialize=0.17,
        doc="HCl cost",  # for 37% sol'n - CatCost v 1.0.4
        units=pyo.units.USD_2020 / pyo.units.kg,
    )
    blk.purity = pyo.Param(
        mutable=True,
        initialize=0.37,
        doc="HCl purity",
        units=pyo.units.dimensionless,
    )

    costing = blk.parent_block()
    costing.register_flow_type("HCl", blk.cost / blk.purity)


def build_naoh_cost_param_block(blk):

    blk.cost = pyo.Param(
        mutable=True,
        initialize=0.59,
        doc="NaOH cost",  # for 30% sol'n - iDST
        units=pyo.units.USD_2020 / pyo.units.kg,
    )

    blk.purity = pyo.Param(
        mutable=True,
        initialize=0.30,
        doc="NaOH purity",
        units=pyo.units.dimensionless,
    )

    costing = blk.parent_block()
    costing.register_flow_type("NaOH", blk.cost / blk.purity)


def build_meoh_cost_param_block(blk):
    # MeOH = Methanol
    blk.cost = pyo.Param(
        mutable=True,
        initialize=3.395,
        doc="MeOH cost",  # for 100% purity - ICIS
        units=pyo.units.USD_2008 / pyo.units.kg,
    )

    blk.purity = pyo.Param(
        mutable=True,
        initialize=1,
        doc="MeOH purity",
        units=pyo.units.dimensionless,
    )

    costing = blk.parent_block()
    costing.register_flow_type("MeOH", blk.cost / blk.purity)


def build_nacl_cost_param_block(blk):

    blk.cost = pyo.Param(
        mutable=True,
        initialize=0.09,
        doc="NaCl cost",  # for solid, 100% purity - CatCost
        units=pyo.units.USD_2020 / pyo.units.kg,
    )

    blk.purity = pyo.Param(
        mutable=True,
        initialize=1,
        doc="NaCl purity",
        units=pyo.units.dimensionless,
    )

    costing = blk.parent_block()
    costing.register_flow_type("NaCl", blk.cost / blk.purity)


def build_ion_exhange_cost_param_block(blk):
    blk.anion_exchange_resin_cost = pyo.Var(
        initialize=205,
        units=pyo.units.USD_2020 / pyo.units.ft**3,
        doc="Anion exchange resin cost per cubic ft. Assumes strong base polystyrenic gel-type Type II. From EPA-WBS cost model.",
    )
    blk.cation_exchange_resin_cost = pyo.Var(
        initialize=153,
        units=pyo.units.USD_2020 / pyo.units.ft**3,
        doc="Cation exchange resin cost per cubic ft. Assumes strong acid polystyrenic gel-type. From EPA-WBS cost model.",
    )
    # Ion exchange pressure vessels costed with power equation, col_vol in gallons:
    #   pressure_vessel_cost = A * col_vol ** b

    blk.vessel_A_coeff = pyo.Var(
        initialize=1596.499333,
        units=pyo.units.USD_2020,
        doc="Ion exchange pressure vessel cost equation - A coeff., Carbon steel w/ stainless steel internals",
    )
    blk.vessel_b_coeff = pyo.Var(
        initialize=0.459496809,
        units=pyo.units.dimensionless,
        doc="Ion exchange pressure vessel cost equation - b coeff., Carbon steel w/ stainless steel internals",
    )

    # Ion exchange backwash/rinse tank costed with power equation, tank_vol in gallons:
    #   bw_tank_cost = A * tank_vol ** b

    blk.backwash_tank_A_coeff = pyo.Var(
        initialize=308.9371309,
        units=pyo.units.USD_2020,
        doc="Ion exchange backwash tank cost equation - A coeff., Steel tank",
    )
    blk.backwash_tank_b_coeff = pyo.Var(
        initialize=0.501467571,
        units=pyo.units.dimensionless,
        doc="Ion exchange backwash tank cost equation - b coeff., Steel tank",
    )
    # Ion exchange regeneration solution tank costed with power equation, tank_vol in gallons:
    #   regen_tank_cost = A * tank_vol ** b

    blk.regen_tank_A_coeff = pyo.Var(
        initialize=57.02158923,
        units=pyo.units.USD_2020,
        doc="Ion exchange regen tank cost equation - A coeff. Stainless steel",
    )
    blk.regen_tank_b_coeff = pyo.Var(
        initialize=0.729325391,
        units=pyo.units.dimensionless,
        doc="Ion exchange regen tank cost equation - b coeff. Stainless steel",
    )
    blk.annual_resin_replacement_factor = pyo.Var(
        initialize=0.05,
        units=pyo.units.year**-1,
        doc="Fraction of ion excange resin replaced per year, 4-5% of bed volume - EPA",
    )
    blk.hazardous_min_cost = pyo.Var(
        initialize=3240,
        units=pyo.units.USD_2020 / pyo.units.year,
        doc="Min cost per hazardous waste shipment - EPA",
    )
    blk.hazardous_resin_disposal = pyo.Var(
        initialize=347.10,
        units=pyo.units.USD_2020 * pyo.units.ton**-1,
        doc="Hazardous resin disposal cost - EPA",
    )
    blk.hazardous_regen_disposal = pyo.Var(
        initialize=3.64,
        units=pyo.units.USD_2020 * pyo.units.gal**-1,
        doc="Hazardous liquid disposal cost - EPA",
    )
    blk.regen_recycle = pyo.Var(
        initialize=1,
        units=pyo.units.dimensionless,
        doc="Number of cycles the regenerant can be reused before disposal",
    )


@register_costing_parameter_block(
    build_rule=build_hcl_cost_param_block,
    parameter_block_name="hcl",
)
@register_costing_parameter_block(
    build_rule=build_naoh_cost_param_block,
    parameter_block_name="naoh",
)
@register_costing_parameter_block(
    build_rule=build_meoh_cost_param_block,
    parameter_block_name="meoh",
)
@register_costing_parameter_block(
    build_rule=build_nacl_cost_param_block,
    parameter_block_name="nacl",
)
@register_costing_parameter_block(
    build_rule=build_ion_exhange_cost_param_block,
    parameter_block_name="ion_exchange",
)
def cost_ion_exchange(blk):
    """
    Volume-based capital cost for Ion Exchange
    """
    make_capital_cost_var(blk)
    make_fixed_operating_cost_var(blk)
    ion_exchange_params = blk.costing_package.ion_exchange
    # Conversions to use units from cost equations in reference
    tot_num_col = blk.unit_model.number_columns + blk.unit_model.number_columns_redund
    col_vol_gal = pyo.units.convert(blk.unit_model.col_vol_per, to_units=pyo.units.gal)
    bed_vol_ft3 = pyo.units.convert(blk.unit_model.bed_vol, to_units=pyo.units.ft**3)

    ix_type = blk.unit_model.ion_exchange_type
    blk.regen_soln_dens = pyo.Param(
        initialize=1000,
        units=pyo.units.kg / pyo.units.m**3,
        mutable=True,
        doc="Density of regeneration solution",
    )
    blk.regen_dose = pyo.Param(
        initialize=300,
        units=pyo.units.kg / pyo.units.m**3,
        mutable=True,
        doc="Regenerant dose required for regeneration per volume of resin [kg regenerant/m3 resin]",
    )
    blk.capital_cost_vessel = pyo.Var(
        initialize=1e5,
        domain=pyo.NonNegativeReals,
        units=blk.costing_package.base_currency,
        doc="Capital cost for one vessel",
    )
    blk.capital_cost_resin = pyo.Var(
        initialize=1e5,
        domain=pyo.NonNegativeReals,
        units=blk.costing_package.base_currency,
        doc="Capital cost for resin for one vessel",
    )
    blk.capital_cost_regen_tank = pyo.Var(
        initialize=1e5,
        domain=pyo.NonNegativeReals,
        units=blk.costing_package.base_currency,
        doc="Capital cost for regeneration solution tank",
    )
    blk.capital_cost_backwash_tank = pyo.Var(
        initialize=1e5,
        domain=pyo.NonNegativeReals,
        units=blk.costing_package.base_currency,
        doc="Capital cost for backwash + rinse solution tank",
    )
    blk.operating_cost_hazardous = pyo.Var(
        initialize=1e5,
        domain=pyo.NonNegativeReals,
        units=blk.costing_package.base_currency / blk.costing_package.base_period,
        doc="Operating cost for hazardous waste disposal",
    )
    blk.flow_mass_regen_soln = pyo.Var(
        initialize=1,
        domain=pyo.NonNegativeReals,
        units=pyo.units.kg / pyo.units.year,
        doc="Regeneration solution flow",
    )
    blk.total_pumping_power = pyo.Var(
        initialize=1,
        domain=pyo.NonNegativeReals,
        units=pyo.units.kilowatt,
        doc="Total pumping power required",
    )

    if ix_type == "cation":
        resin_cost = ion_exchange_params.cation_exchange_resin_cost

    elif ix_type == "anion":
        resin_cost = ion_exchange_params.anion_exchange_resin_cost

    blk.capital_cost_vessel_constraint = pyo.Constraint(
        expr=blk.capital_cost_vessel
        == pyo.units.convert(
            (
                ion_exchange_params.vessel_A_coeff
                * (col_vol_gal / pyo.units.gallon) ** ion_exchange_params.vessel_b_coeff
            ),
            to_units=blk.costing_package.base_currency,
        )
    )
    blk.capital_cost_resin_constraint = pyo.Constraint(
        expr=blk.capital_cost_resin
        == pyo.units.convert(
            resin_cost * bed_vol_ft3, to_units=blk.costing_package.base_currency
        )
    )
    if blk.unit_model.config.regenerant == "single_use":
        blk.capital_cost_regen_tank.fix(0)
        blk.flow_mass_regen_soln.fix(0)
        blk.flow_vol_resin = pyo.Var(
            initialize=1e5,
            bounds=(0, None),
            units=pyo.units.m**3 / blk.costing_package.base_period,
            doc="Volumetric flow of resin per cycle",  # assumes you are only replacing the operational columns, t_cycle = t_breakthru
        )
        blk.single_use_resin_replacement_cost = pyo.Var(
            initialize=1e5,
            bounds=(0, None),
            units=blk.costing_package.base_currency / blk.costing_package.base_period,
            doc="Operating cost for using single-use resin (i.e., no regeneration)",
        )

        blk.flow_vol_resin_constraint = pyo.Constraint(
            expr=blk.flow_vol_resin
            == pyo.units.convert(
                blk.unit_model.bed_vol_tot / blk.unit_model.t_breakthru,
                to_units=pyo.units.m**3 / blk.costing_package.base_period,
            )
        )
        blk.mass_flow_resin = pyo.units.convert(
            blk.flow_vol_resin * blk.unit_model.resin_bulk_dens,
            to_units=pyo.units.ton / blk.costing_package.base_period,
        )
    else:
        blk.backwash_tank_vol = pyo.Expression(
            expr=pyo.units.convert(
                (
                    blk.unit_model.bw_flow * blk.unit_model.t_bw
                    + blk.unit_model.rinse_flow * blk.unit_model.t_rinse
                ),
                to_units=pyo.units.gal,
            )
        )

        blk.regeneration_tank_vol = pyo.Expression(
            expr=pyo.units.convert(
                blk.unit_model.regen_tank_vol,
                to_units=pyo.units.gal,
            )
        )
        blk.capital_cost_backwash_tank_constraint = pyo.Constraint(
            expr=blk.capital_cost_backwash_tank
            == pyo.units.convert(
                ion_exchange_params.backwash_tank_A_coeff
                * (blk.backwash_tank_vol / pyo.units.gallon)
                ** ion_exchange_params.backwash_tank_b_coeff,
                to_units=blk.costing_package.base_currency,
            )
        )
        blk.capital_cost_regen_tank_constraint = pyo.Constraint(
            expr=blk.capital_cost_regen_tank
            == pyo.units.convert(
                ion_exchange_params.regen_tank_A_coeff
                * (blk.regeneration_tank_vol / pyo.units.gallon)
                ** ion_exchange_params.regen_tank_b_coeff,
                to_units=blk.costing_package.base_currency,
            )
        )

    blk.costing_package.add_cost_factor(blk, "TIC")
    blk.capital_cost_constraint = pyo.Constraint(
        expr=blk.capital_cost
        == blk.cost_factor
        * pyo.units.convert(
            (
                ((blk.capital_cost_vessel + blk.capital_cost_resin) * tot_num_col)
                + blk.capital_cost_backwash_tank
                + blk.capital_cost_regen_tank
            ),
            to_units=blk.costing_package.base_currency,
        )
    )
    if blk.unit_model.config.hazardous_waste:

        if blk.unit_model.config.regenerant == "single_use":
            blk.operating_cost_hazardous_constraint = pyo.Constraint(
                expr=blk.operating_cost_hazardous
                == pyo.units.convert(
                    (blk.mass_flow_resin * ion_exchange_params.hazardous_resin_disposal)
                    + ion_exchange_params.hazardous_min_cost,
                    to_units=blk.costing_package.base_currency
                    / blk.costing_package.base_period,
                )
            )
        else:
            bed_mass_ton = pyo.units.convert(
                blk.unit_model.bed_vol * blk.unit_model.resin_bulk_dens,
                to_units=pyo.units.ton,
            )
            blk.operating_cost_hazardous_constraint = pyo.Constraint(
                expr=blk.operating_cost_hazardous
                == pyo.units.convert(
                    (
                        bed_mass_ton
                        * tot_num_col
                        * ion_exchange_params.hazardous_resin_disposal
                    )
                    * ion_exchange_params.annual_resin_replacement_factor
                    + pyo.units.convert(
                        blk.flow_mass_regen_soln / blk.regen_soln_dens,
                        to_units=pyo.units.gal / pyo.units.year,
                    )
                    * ion_exchange_params.hazardous_regen_disposal
                    + ion_exchange_params.hazardous_min_cost,
                    to_units=blk.costing_package.base_currency
                    / blk.costing_package.base_period,
                )
            )
    else:
        blk.operating_cost_hazardous.fix(0)
    if blk.unit_model.config.regenerant == "single_use":

        blk.single_use_resin_replacement_cost_constraint = pyo.Constraint(
            expr=blk.single_use_resin_replacement_cost
            == pyo.units.convert(
                blk.flow_vol_resin * resin_cost,
                to_units=blk.costing_package.base_currency
                / blk.costing_package.base_period,
            )
        )

        blk.fixed_operating_cost_constraint = pyo.Constraint(
            expr=blk.fixed_operating_cost
            == blk.single_use_resin_replacement_cost + blk.operating_cost_hazardous
        )

    else:
        blk.fixed_operating_cost_constraint = pyo.Constraint(
            expr=blk.fixed_operating_cost
            == pyo.units.convert(
                (
                    (
                        bed_vol_ft3
                        * tot_num_col
                        * ion_exchange_params.annual_resin_replacement_factor
                        * resin_cost
                    )
                ),
                to_units=blk.costing_package.base_currency
                / blk.costing_package.base_period,
            )
            + blk.operating_cost_hazardous
        )

        blk.flow_mass_regen_soln_constraint = pyo.Constraint(
            expr=blk.flow_mass_regen_soln
            == pyo.units.convert(
                (
                    (blk.regen_dose * blk.unit_model.bed_vol * tot_num_col)
                    / (blk.unit_model.t_cycle)
                )
                / ion_exchange_params.regen_recycle,
                to_units=pyo.units.kg / pyo.units.year,
            )
        )

        blk.costing_package.cost_flow(
            blk.flow_mass_regen_soln, blk.unit_model.config.regenerant
        )

    if blk.unit_model.config.regenerant == "single_use":
        power_expr = blk.unit_model.main_pump_power
    else:
        power_expr = (
            blk.unit_model.main_pump_power
            + blk.unit_model.regen_pump_power
            + blk.unit_model.bw_pump_power
            + blk.unit_model.rinse_pump_power
        )

    blk.total_pumping_power_constr = pyo.Constraint(
        expr=blk.total_pumping_power == power_expr
    )

    blk.costing_package.cost_flow(blk.total_pumping_power, "electricity")



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
This module contains a zero-order representation of a CANDO+P reactor unit.
"""

import pyomo.environ as pyo
from pyomo.environ import Var, units as pyunits
from idaes.core import declare_process_block_class
from watertap.core import build_sido_reactive, ZeroOrderBaseData

# Some more information about this module
__author__ = "Travis Arnold"


@declare_process_block_class("CANDOPZO")
class CANDOPData(ZeroOrderBaseData):
    """
    Zero-Order model for a CANDO+P reactor unit.
    """

    CONFIG = ZeroOrderBaseData.CONFIG()

    def build(self):
        super().build()

        self._tech_type = "CANDO_P"

        build_sido_reactive(self)

        # Create electricity variable
        self.electricity = Var(
            self.flowsheet().time,
            units=pyunits.kW,
            bounds=(0, None),
            doc="Electricity consumption of unit",
        )
        self._perf_var_dict["Electricity Demand"] = self.electricity

        # Create electricity intensity variable and constraint. For this
        # model, electricity demand is calculated based on the amount of
        # nitrogen reacted.
        # TODO The information I have says that this electricity consumption
        #      accounts pumping, oxygenation, and stirring. At some point,
        #      perhaps we should come back and adjust the model to account for
        #      pumping costs separately.
        self.electricity_intensity_N = Var(
            units=pyunits.kWh / pyunits.kg,
            bounds=(0, None),
            doc="Electricity demand per kg N reacted",
        )
        self._fixed_perf_vars.append(self.electricity_intensity_N)
        self._perf_var_dict["Electricity Intensity"] = self.electricity_intensity_N

        @self.Constraint(
            self.flowsheet().time,
            doc="Constraint for electricity consumption based on " "nitrogen consumed.",
        )
        def electricity_consumption(b, t):
            return b.electricity[t] == (
                pyunits.convert(
                    b.extent_of_reaction[t, "n_reaction"] * b.electricity_intensity_N,
                    to_units=pyunits.kW,
                )
            )

        # Create oxygen demand variables and constraint. The amount of oxygen
        # consumed is assumed to be a linear function of the amount of
        # nitrogen reacted.
        self.O2_demand = Var(
            self.flowsheet().time,
            units=pyunits.kg / pyunits.s,
            bounds=(0, None),
            doc="Oxygen demand",
        )
        self._perf_var_dict["Oxygen Demand"] = self.O2_demand
        self.oxygen_nitrogen_ratio = Var(
            units=pyunits.dimensionless,
            bounds=(0, None),
            doc="Oxygen consumed - nitrogen reacted ratio",
        )
        self._fixed_perf_vars.append(self.oxygen_nitrogen_ratio)
        self._perf_var_dict[
            "Oxygen consumed / nitrogen reacted ratio (mass basis)"
        ] = self.oxygen_nitrogen_ratio

        @self.Constraint(
            self.flowsheet().time, doc="Constraint for oxygen consumption."
        )
        def oxygen_consumption(b, t):
            return b.O2_demand[t] == (
                b.extent_of_reaction[t, "n_reaction"] * b.oxygen_nitrogen_ratio
            )

    @property
    def default_costing_method(self):
        return self.cost_CANDOP

    @staticmethod
    def cost_CANDOP(blk):
        """
        General method for costing CANDO+P reactor.
        """
        t0 = blk.flowsheet().time.first()

        # Get parameter dict from database
        parameter_dict = blk.unit_model.config.database.get_unit_operation_parameters(
            blk.unit_model._tech_type, subtype=blk.unit_model.config.process_subtype
        )

        # Get costing parameter sub-block for this technology
        size_param, size_cost = blk.unit_model._get_tech_parameters(
            blk,
            parameter_dict,
            blk.unit_model.config.process_subtype,
            ["sizing_parameter", "sizing_cost"],
        )

        # Add cost variable and constraint
        blk.capital_cost = pyo.Var(
            initialize=1,
            units=blk.config.flowsheet_costing_block.base_currency,
            bounds=(0, None),
            doc="Capital cost of unit operation",
        )

        expr = pyo.units.convert(
            blk.unit_model.properties_in[t0].flow_vol * size_param * size_cost,
            to_units=blk.config.flowsheet_costing_block.base_currency,
        )

        # Determine if a costing factor is required
        blk.costing_package.add_cost_factor(
            blk, parameter_dict["capital_cost"]["cost_factor"]
        )

        blk.capital_cost_constraint = pyo.Constraint(
            expr=blk.capital_cost == blk.cost_factor * expr
        )

        # Register flows
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.electricity[t0], "electricity"
        )



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Tests for CSTR unit model with injection.
Authors: Andrew Lee, Vibhav Dabadghao
"""

import pytest
from pyomo.environ import (
    check_optimal_termination,
    ConcreteModel,
    units,
    value,
    assert_optimal_termination,
)
from idaes.core import (
    FlowsheetBlock,
    MaterialBalanceType,
    EnergyBalanceType,
    MomentumBalanceType,
)
from idaes.models.properties.examples.saponification_thermo import (
    SaponificationParameterBlock,
)
from idaes.models.properties.examples.saponification_reactions import (
    SaponificationReactionParameterBlock,
)
from idaes.core.util.model_statistics import (
    degrees_of_freedom,
    number_variables,
    number_total_constraints,
    number_unused_variables,
)
from idaes.core.util.testing import (
    PhysicalParameterTestBlock,
    ReactionParameterTestBlock,
    initialization_tester,
)
from idaes.core.solvers import get_solver
from pyomo.util.check_units import assert_units_consistent, assert_units_equivalent

from watertap.unit_models.cstr_injection import CSTR_Injection
from idaes.core import UnitModelCostingBlock
from watertap.costing import WaterTAPCosting


# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()


# -----------------------------------------------------------------------------
@pytest.mark.unit
def test_config():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.properties = PhysicalParameterTestBlock()
    m.fs.reactions = ReactionParameterTestBlock(property_package=m.fs.properties)

    m.fs.unit = CSTR_Injection(
        property_package=m.fs.properties, reaction_package=m.fs.reactions
    )

    # Check unit config arguments
    assert len(m.fs.unit.config) == 14

    assert m.fs.unit.config.material_balance_type == MaterialBalanceType.useDefault
    assert m.fs.unit.config.energy_balance_type == EnergyBalanceType.useDefault
    assert m.fs.unit.config.momentum_balance_type == MomentumBalanceType.pressureTotal
    assert not m.fs.unit.config.has_heat_transfer
    assert not m.fs.unit.config.has_pressure_change
    assert not m.fs.unit.config.has_equilibrium_reactions
    assert not m.fs.unit.config.has_phase_equilibrium
    assert not m.fs.unit.config.has_heat_of_reaction
    assert m.fs.unit.config.property_package is m.fs.properties
    assert m.fs.unit.config.reaction_package is m.fs.reactions


# -----------------------------------------------------------------------------
class TestSaponification(object):
    @pytest.fixture(scope="class")
    def sapon(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.properties = SaponificationParameterBlock()
        m.fs.reactions = SaponificationReactionParameterBlock(
            property_package=m.fs.properties
        )

        m.fs.unit = CSTR_Injection(
            property_package=m.fs.properties,
            reaction_package=m.fs.reactions,
            has_equilibrium_reactions=False,
            has_heat_transfer=True,
            has_heat_of_reaction=True,
            has_pressure_change=True,
        )

        m.fs.unit.inlet.flow_vol.fix(1.0e-03)
        m.fs.unit.inlet.conc_mol_comp[0, "H2O"].fix(55388.0)
        m.fs.unit.inlet.conc_mol_comp[0, "NaOH"].fix(100.0)
        m.fs.unit.inlet.conc_mol_comp[0, "EthylAcetate"].fix(100.0)
        m.fs.unit.inlet.conc_mol_comp[0, "SodiumAcetate"].fix(0.0)
        m.fs.unit.inlet.conc_mol_comp[0, "Ethanol"].fix(0.0)

        m.fs.unit.inlet.temperature.fix(303.15)
        m.fs.unit.inlet.pressure.fix(101325.0)

        m.fs.unit.volume.fix(1.5e-03)
        m.fs.unit.heat_duty.fix(0)
        m.fs.unit.deltaP.fix(0)
        m.fs.unit.injection.fix(0)

        return m

    @pytest.mark.build
    @pytest.mark.unit
    def test_build(self, sapon):
        assert hasattr(sapon.fs.unit, "inlet")
        assert len(sapon.fs.unit.inlet.vars) == 4
        assert hasattr(sapon.fs.unit.inlet, "flow_vol")
        assert hasattr(sapon.fs.unit.inlet, "conc_mol_comp")
        assert hasattr(sapon.fs.unit.inlet, "temperature")
        assert hasattr(sapon.fs.unit.inlet, "pressure")

        assert hasattr(sapon.fs.unit, "outlet")
        assert len(sapon.fs.unit.outlet.vars) == 4
        assert hasattr(sapon.fs.unit.outlet, "flow_vol")
        assert hasattr(sapon.fs.unit.outlet, "conc_mol_comp")
        assert hasattr(sapon.fs.unit.outlet, "temperature")
        assert hasattr(sapon.fs.unit.outlet, "pressure")

        assert hasattr(sapon.fs.unit, "cstr_performance_eqn")
        assert hasattr(sapon.fs.unit, "volume")
        assert hasattr(sapon.fs.unit, "hydraulic_retention_time")
        assert hasattr(sapon.fs.unit, "heat_duty")
        assert hasattr(sapon.fs.unit, "deltaP")

        assert hasattr(sapon.fs.unit, "injection")
        for k in sapon.fs.unit.injection:
            assert (
                sapon.fs.unit.injection[k]
                == sapon.fs.unit.control_volume.mass_transfer_term[k]
            )

        assert number_variables(sapon) == 34
        assert number_total_constraints(sapon) == 18
        assert number_unused_variables(sapon) == 0

    @pytest.mark.component
    def test_units(self, sapon):
        assert_units_consistent(sapon)
        assert_units_equivalent(sapon.fs.unit.volume[0], units.m**3)
        assert_units_equivalent(sapon.fs.unit.heat_duty[0], units.W)
        assert_units_equivalent(sapon.fs.unit.deltaP[0], units.Pa)

    @pytest.mark.unit
    def test_dof(self, sapon):
        assert degrees_of_freedom(sapon) == 0

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, sapon):
        initialization_tester(sapon)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, sapon):
        results = solver.solve(sapon)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, sapon):
        assert pytest.approx(101325.0, abs=1e-2) == value(
            sapon.fs.unit.outlet.pressure[0]
        )
        assert pytest.approx(304.09, abs=1e-2) == value(
            sapon.fs.unit.outlet.temperature[0]
        )
        assert pytest.approx(20.32, abs=1e-2) == value(
            sapon.fs.unit.outlet.conc_mol_comp[0, "EthylAcetate"]
        )
        assert pytest.approx(0.03888, abs=1e-3) == value(
            sapon.fs.unit.electricity_consumption[0]
        )
        assert pytest.approx(1.5, abs=1e-3) == value(
            sapon.fs.unit.hydraulic_retention_time[0]
        )

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, sapon):
        assert (
            abs(
                value(
                    sapon.fs.unit.inlet.flow_vol[0] - sapon.fs.unit.outlet.flow_vol[0]
                )
            )
            <= 1e-6
        )
        assert (
            abs(
                value(
                    sapon.fs.unit.inlet.flow_vol[0]
                    * sum(
                        sapon.fs.unit.inlet.conc_mol_comp[0, j]
                        for j in sapon.fs.properties.component_list
                    )
                    - sapon.fs.unit.outlet.flow_vol[0]
                    * sum(
                        sapon.fs.unit.outlet.conc_mol_comp[0, j]
                        for j in sapon.fs.properties.component_list
                    )
                )
            )
            <= 1e-6
        )

        assert pytest.approx(3904.51, abs=1e-2) == value(
            sapon.fs.unit.control_volume.heat_of_reaction[0]
        )
        assert (
            abs(
                value(
                    (
                        sapon.fs.unit.inlet.flow_vol[0]
                        * sapon.fs.properties.dens_mol
                        * sapon.fs.properties.cp_mol
                        * (
                            sapon.fs.unit.inlet.temperature[0]
                            - sapon.fs.properties.temperature_ref
                        )
                    )
                    - (
                        sapon.fs.unit.outlet.flow_vol[0]
                        * sapon.fs.properties.dens_mol
                        * sapon.fs.properties.cp_mol
                        * (
                            sapon.fs.unit.outlet.temperature[0]
                            - sapon.fs.properties.temperature_ref
                        )
                    )
                    + sapon.fs.unit.control_volume.heat_of_reaction[0]
                )
            )
            <= 1e-3
        )

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_costing(self, sapon):
        m = sapon

        m.fs.costing = WaterTAPCosting()

        m.fs.unit.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)
        m.fs.costing.cost_process()
        m.fs.costing.add_LCOW(m.fs.unit.control_volume.properties_out[0].flow_vol)
        solver = get_solver()
        results = solver.solve(m)

        assert_optimal_termination(results)

        # Check solutions
        assert pytest.approx(7.75429 * 2, rel=1e-5) == value(
            m.fs.unit.costing.capital_cost
        )
        assert pytest.approx(0.00082698, rel=1e-5) == value(m.fs.costing.LCOW)

    @pytest.mark.unit
    def test_get_performance_contents(self, sapon):
        perf_dict = sapon.fs.unit._get_performance_contents()

        assert perf_dict == {
            "vars": {
                "Volume": sapon.fs.unit.volume[0],
                "Injection [('Liq', 'H2O')]": sapon.fs.unit.injection[0, "Liq", "H2O"],
                "Injection [('Liq', 'NaOH')]": sapon.fs.unit.injection[
                    0, "Liq", "NaOH"
                ],
                "Injection [('Liq', 'EthylAcetate')]": sapon.fs.unit.injection[
                    0, "Liq", "EthylAcetate"
                ],
                "Injection [('Liq', 'SodiumAcetate')]": sapon.fs.unit.injection[
                    0, "Liq", "SodiumAcetate"
                ],
                "Injection [('Liq', 'Ethanol')]": sapon.fs.unit.injection[
                    0, "Liq", "Ethanol"
                ],
                "Heat Duty": sapon.fs.unit.heat_duty[0],
                "Pressure Change": sapon.fs.unit.deltaP[0],
                "Electricity Consumption": sapon.fs.unit.electricity_consumption[0],
            }
        }



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

from pyomo.environ import (
    Block,
    Constraint,
    Expression,
    Var,
    Param,
    value,
)
from idaes.core import UnitModelCostingBlock
from watertap.costing import (
    WaterTAPCosting,
    PumpType,
    EnergyRecoveryDeviceType,
    MixerType,
    ROType,
)

import idaes.core.util.scaling as iscale


def build_costing(m, costing_package=WaterTAPCosting, **kwargs):
    """Add costing to a given flowsheet

    Args:
        m: model
        costing_package : FlowsheetCostingBlock
    """

    # call get_costing for each unit model
    m.fs.costing = costing_package()
    # the full_treatment_train uses a lower than default value
    # for factor_maintenance_labor_chemical
    m.fs.costing.factor_maintenance_labor_chemical.fix(0.02)
    crf = m.fs.costing.factor_capital_annualization

    # Nanofiltration
    if hasattr(m.fs, "NF"):
        if kwargs["NF_type"] == "ZO":
            m.fs.NF.costing = UnitModelCostingBlock(
                flowsheet_costing_block=m.fs.costing
            )
        elif kwargs["NF_type"] == "Sep":
            raise NotImplementedError(
                "get_costing will not be implemented for the NF separator model."
            )
    if hasattr(m.fs, "pump_NF"):
        m.fs.pump_NF.costing = UnitModelCostingBlock(
            flowsheet_costing_block=m.fs.costing,
            costing_method_arguments={"pump_type": PumpType.low_pressure},
        )

    # Reverse Osmosis
    if hasattr(m.fs, "RO"):
        if kwargs["RO_type"] == "0D" or kwargs["RO_type"] == "1D":
            m.fs.RO.costing = UnitModelCostingBlock(
                flowsheet_costing_block=m.fs.costing
            )
        elif kwargs["RO_type"] == "Sep":
            raise NotImplementedError(
                "get_costing will not be implemented for the RO separator model."
            )

    # Stage 2 RO
    if hasattr(m.fs, "RO2"):
        m.fs.RO2.costing = UnitModelCostingBlock(
            flowsheet_costing_block=m.fs.costing,
            costing_method_arguments={"ro_type": ROType.high_pressure},
        )

    # Pump
    if hasattr(m.fs, "pump_RO"):
        m.fs.pump_RO.costing = UnitModelCostingBlock(
            flowsheet_costing_block=m.fs.costing,
            costing_method_arguments={"pump_type": PumpType.high_pressure},
        )

    # Stage 2 pump
    if hasattr(m.fs, "pump_RO2"):
        m.fs.pump_RO2.costing = UnitModelCostingBlock(
            flowsheet_costing_block=m.fs.costing,
            costing_method_arguments={"pump_type": PumpType.high_pressure},
        )

    # ERD
    if hasattr(m.fs, "ERD"):
        m.fs.ERD.costing = UnitModelCostingBlock(
            flowsheet_costing_block=m.fs.costing,
            costing_method_arguments={
                "energy_recovery_device_type": EnergyRecoveryDeviceType.pressure_exchanger
            },
        )

    # Pretreatment
    if hasattr(m.fs, "stoich_softening_mixer_unit"):
        m.fs.stoich_softening_mixer_unit.costing = UnitModelCostingBlock(
            flowsheet_costing_block=m.fs.costing,
            costing_method_arguments={
                "mixer_type": MixerType.CaOH2,
                "dosing_rate": m.fs.stoich_softening_mixer_unit.dosing_rate,
            },
        )

    # Post-treatment
    if hasattr(m.fs, "ideal_naocl_mixer_unit"):
        # print('FOUND CHLORINATION UNIT')
        m.fs.ideal_naocl_mixer_unit.costing = UnitModelCostingBlock(
            flowsheet_costing_block=m.fs.costing,
            costing_method_arguments={
                "mixer_type": MixerType.NaOCl,
                "dosing_rate": m.fs.ideal_naocl_mixer_unit.dosing_rate,
            },
        )

    if hasattr(m.fs, "mixer_permeate"):
        m.fs.mixer_permeate.costing = UnitModelCostingBlock(
            flowsheet_costing_block=m.fs.costing,
            costing_method_arguments={"mixer_type": MixerType.default},
        )

    # call get_system_costing for whole flowsheet
    m.fs.costing.cost_process()
    m.fs.costing.add_annual_water_production(m.fs.treated_flow_vol)
    m.fs.costing.add_LCOW(m.fs.treated_flow_vol)

    if hasattr(m.fs, "stoich_softening_mixer_unit"):
        m.fs.lime_softening_unit_capex = Expression(
            expr=m.fs.stoich_softening_mixer_unit.costing.capital_cost
            / m.fs.costing.annual_water_production
            * crf
        )
    else:
        m.fs.lime_softening_unit_capex = Expression(expr=0)

    if hasattr(m.fs, "ideal_naocl_mixer_unit"):
        m.fs.chlorination_unit_capex = Expression(
            expr=m.fs.ideal_naocl_mixer_unit.costing.capital_cost
            / m.fs.costing.annual_water_production
            * crf
        )
    else:
        m.fs.chlorination_unit_capex = Expression(expr=0)

    # apply scaling to cost variables and constraints
    scale_costing(m)


def scale_costing(self):
    for b_unit in self.component_objects(Block, descend_into=True):
        if hasattr(b_unit, "costing"):
            base = b_unit.costing
            for var in base.component_objects(Var):
                if iscale.get_scaling_factor(var) is None:
                    iscale.set_scaling_factor(var, 1e-3)
            for con in base.component_data_objects(Constraint):
                iscale.constraint_scaling_transform(con, 1e-3)


def display_costing(m):
    crf = m.fs.costing.factor_capital_annualization
    if not hasattr(m.fs, "pump_RO2"):
        m.fs.pump_RO2 = Block()
        m.fs.pump_RO2.costing = Block()
        m.fs.pump_RO2.costing.fixed_operating_cost = Param(initialize=0)
    if not hasattr(m.fs, "NF"):
        m.fs.NF = Block()
        m.fs.NF.costing = Block()
        m.fs.NF.costing.fixed_operating_cost = Param(initialize=0)
    if not hasattr(m.fs, "RO2"):
        m.fs.RO2 = Block()
        m.fs.RO2.costing = Block()
        m.fs.RO2.costing.fixed_operating_cost = Param(initialize=0)

    # UNITS FOR ALL COST COMPONENTS [=] $/m3 of permeate water produced
    cost_dict = {
        "LCOW": m.fs.costing.LCOW,  # Total LCOW
        "Total CAPEX": m.fs.costing.total_capital_cost
        * crf
        / m.fs.costing.annual_water_production,  # Direct + Indirect CAPEX
        "Direct CAPEX": m.fs.costing.aggregate_capital_cost
        * crf
        / m.fs.costing.annual_water_production,  # Direct CAPEX for all system components
        "Indirect CAPEX": (
            m.fs.costing.total_capital_cost - m.fs.costing.aggregate_capital_cost
        )
        * crf
        / m.fs.costing.annual_water_production,  # Indirect CAPEX for miscellaneous items
        "Total OPEX": m.fs.costing.total_operating_cost
        / m.fs.costing.annual_water_production,  # Total OPEX
        "Labor & Maintenance Costs": m.fs.costing.maintenance_labor_chemical_operating_cost
        / m.fs.costing.annual_water_production,
        "Total Electricity Cost": m.fs.costing.aggregate_flow_costs["electricity"]
        / m.fs.costing.annual_water_production
        if "electricity" in m.fs.costing.aggregate_flow_costs
        else 0.0,
        "Total Membrane Replacement Cost": (
            m.fs.NF.costing.fixed_operating_cost
            + m.fs.RO.costing.fixed_operating_cost
            + m.fs.RO2.costing.fixed_operating_cost
        )
        / m.fs.costing.annual_water_production,
        "Lime softener CAPEX": m.fs.lime_softening_unit_capex,
        "Lime softener OPEX": m.fs.costing.aggregate_flow_costs["CaOH2"]
        / m.fs.costing.annual_water_production
        if "CaOH2" in m.fs.costing.aggregate_flow_costs
        else 0.0,
        "Chlorination CAPEX": m.fs.chlorination_unit_capex,
        "Chlorination OPEX": m.fs.costing.aggregate_flow_costs["NaOCl"]
        / m.fs.costing.annual_water_production
        if "NaOCl" in m.fs.costing.aggregate_flow_costs
        else 0.0,
    }

    for item, val in cost_dict.items():
        print(f"{item} = {value(val)}")

    return cost_dict



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

"""
This module contains a zero-order representation of a peracetic acid (PAA) water
disinfection unit.
"""

import pyomo.environ as pyo
from pyomo.environ import Var, units as pyunits
from idaes.core import declare_process_block_class
from watertap.core import build_sido_reactive, constant_intensity, ZeroOrderBaseData


# Some more information about this module
__author__ = "Travis Arnold"


@declare_process_block_class("PeraceticAcidDisinfectionZO")
class PeraceticAcidDisinfectionData(ZeroOrderBaseData):
    """
    Zero-Order model for a peracetic acid water disinfection unit.
    """

    CONFIG = ZeroOrderBaseData.CONFIG()

    def build(self):
        super().build()

        self._tech_type = "peracetic_acid_disinfection"

        build_sido_reactive(self)
        constant_intensity(self)

        # Create hydraulic retention time variable
        self.HRT = Var(
            units=pyunits.hr,
            bounds=(0, None),
            doc="Hydraulic retention time of water treatment unit",
        )
        self._perf_var_dict["Hydraulic Retention Time"] = self.HRT
        self._fixed_perf_vars.append(self.HRT)

        # Create variable for mass of an E. coli cell
        self.ecoli_cell_mass = Var(
            units=pyunits.kg,
            bounds=(0, None),
            doc="Average mass of an E. coli cell",
        )
        self._perf_var_dict["E. coli Average Cell Mass"] = self.ecoli_cell_mass
        self._fixed_perf_vars.append(self.ecoli_cell_mass)

        # Create variable for weight fraction of PAA in disinfection solution
        self.disinfection_solution_wt_frac_PAA = Var(
            units=pyunits.dimensionless,
            bounds=(0, 1),
            doc="Weight fraction of PAA in disinfection solution",
        )
        self._perf_var_dict[
            "Weight fraction PAA in disinfection solution"
        ] = self.disinfection_solution_wt_frac_PAA
        self._fixed_perf_vars.append(self.disinfection_solution_wt_frac_PAA)

        # Create variable for disinfection solution density
        self.disinfection_solution_density = Var(
            units=pyunits.kg / pyunits.liter,
            bounds=(0, None),
            doc="Disinfection solution density",
        )
        self._perf_var_dict[
            "disinfection solution density"
        ] = self.disinfection_solution_density
        self._fixed_perf_vars.append(self.disinfection_solution_density)

        # Create variable for disinfection solution volumetric flowrate
        self.disinfection_solution_flow_vol = Var(
            self.flowsheet().time,
            units=pyunits.L / pyunits.s,
            bounds=(0, None),
            doc="Volumetric flowrate of disinfection solution",
        )
        self._perf_var_dict[
            "Disinfection solution volumetric flowrate"
        ] = self.disinfection_solution_flow_vol

        # Create constraint to calculate disinfection solution flowrate
        @self.Constraint(
            self.flowsheet().time, doc="Constraint for disinfection solution flowrate"
        )
        def disinfection_solution_flow_vol_rule(b, t):
            return pyunits.convert(
                b.inlet.flow_mass_comp[t, "peracetic_acid"],
                to_units=pyunits.kg / pyunits.s,
            ) == pyunits.convert(
                b.disinfection_solution_flow_vol[t]
                * b.disinfection_solution_density
                * b.disinfection_solution_wt_frac_PAA,
                to_units=pyunits.kg / pyunits.s,
            )

        # Create reactor volume variable
        self.reactor_volume = Var(
            units=pyunits.m**3,
            bounds=(0, None),
            doc="Volume of water treatment unit",
        )
        self._perf_var_dict["Reactor Volume"] = self.reactor_volume

        # Create constraint relating HRT, volume, and volumetric flowrate
        @self.Constraint(self.flowsheet().time, doc="Constraint for reactor volume")
        def reactor_volume_rule(b, t):
            return b.reactor_volume == (
                pyunits.convert(
                    b.HRT * b.properties_in[t].flow_vol, to_units=pyunits.m**3
                )
            )

        # Create variable for E. coli concentration at reactor inlet
        self.inlet_ecoli_conc = Var(
            self.flowsheet().time,
            units=pyunits.liter**-1,
            bounds=(0, None),
            doc="Concentration of E. coli at reactor inlet",
        )
        self._perf_var_dict["Inlet E. coli Concentration"] = self.inlet_ecoli_conc

        # Create constraint relating E. coli inlet mass flow rate and
        # concentration
        @self.Constraint(
            self.flowsheet().time, doc="Constraint for E. coli inlet concentration"
        )
        def ecoli_inlet_concentration_rule(b, t):
            return pyunits.convert(
                b.inlet.flow_mass_comp[t, "total_coliforms_fecal_ecoli"],
                to_units=pyunits.kg / pyunits.s,
            ) == pyunits.convert(
                b.inlet_ecoli_conc[t] * b.ecoli_cell_mass * b.properties_in[t].flow_vol,
                to_units=pyunits.kg / pyunits.s,
            )

        # Create variable for E. coli concentration at reactor outlet
        self.outlet_ecoli_conc = Var(
            self.flowsheet().time,
            units=pyunits.liter**-1,
            bounds=(0, None),
            doc="Concentration of E. coli at reactor outlet",
        )
        self._perf_var_dict["Outlet E. coli Concentration"] = self.outlet_ecoli_conc

        # Create constraint relating E. coli outlet mass flow rate and
        # concentration
        @self.Constraint(
            self.flowsheet().time, doc="Constraint for E. coli outlet concentration"
        )
        def ecoli_outlet_concentration_rule(b, t):
            return pyunits.convert(
                b.treated.flow_mass_comp[t, "total_coliforms_fecal_ecoli"],
                to_units=pyunits.kg / pyunits.s,
            ) == pyunits.convert(
                b.outlet_ecoli_conc[t]
                * b.ecoli_cell_mass
                * b.properties_treated[t].flow_vol,
                to_units=pyunits.kg / pyunits.s,
            )

    @property
    def default_costing_method(self):
        return self.cost_peracetic_acid

    @staticmethod
    def cost_peracetic_acid(blk):
        """
        General method for costing peracetic acid water disinfection.
        """
        t0 = blk.flowsheet().time.first()

        # Get parameter dict from database
        parameter_dict = blk.unit_model.config.database.get_unit_operation_parameters(
            blk.unit_model._tech_type,
            subtype=blk.unit_model.config.process_subtype,
        )

        # Get costing parameter sub-block for this technology
        sizing_cost = blk.unit_model._get_tech_parameters(
            blk,
            parameter_dict,
            blk.unit_model.config.process_subtype,
            ["sizing_cost"],
        )

        # Add cost variable and constraint
        blk.capital_cost = pyo.Var(
            initialize=1,
            units=blk.config.flowsheet_costing_block.base_currency,
            bounds=(0, None),
            doc="Capital cost of unit operation",
        )

        expr = pyo.units.convert(
            blk.unit_model.properties_in[t0].flow_vol * sizing_cost,
            to_units=blk.config.flowsheet_costing_block.base_currency,
        )

        # Determine if a costing factor is required
        blk.costing_package.add_cost_factor(
            blk, parameter_dict["capital_cost"]["cost_factor"]
        )

        blk.capital_cost_constraint = pyo.Constraint(
            expr=blk.capital_cost == blk.cost_factor * expr
        )

        # Register flows
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.electricity[t0], "electricity"
        )
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.disinfection_solution_flow_vol[t0], "disinfection_solution"
        )



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Tests for ADM1 reaction package.

Verified against results from:

X. Flores-Alsina, K. Solon, C.K. Mbamba, S. Tait, K.V. Gernaey, U. Jeppsson, D.J. Batstone,
Modelling phosphorus (P), sulfur (S) and iron (Fe) interactions for dynamic simulations of anaerobic digestion processes,
Water Research. 95 (2016) 370-382. https://www.sciencedirect.com/science/article/pii/S0043135416301397

Authors: Chenyu Wang, Marcus Holly, Adam Atia, Xinhong Liu
"""

import pytest

from pyomo.environ import (
    assert_optimal_termination,
    ConcreteModel,
    Param,
    log10,
    value,
    Var,
    Constraint,
)
from pyomo.util.check_units import assert_units_consistent
from idaes.core import FlowsheetBlock
from watertap.unit_models.anaerobic_digestor import AD
from idaes.core import MaterialFlowBasis
from idaes.core.solvers import get_solver
import idaes.core.util.scaling as iscale
from idaes.core.util.model_statistics import degrees_of_freedom
from watertap.property_models.anaerobic_digestion.modified_adm1_properties import (
    ModifiedADM1ParameterBlock,
)
from watertap.property_models.anaerobic_digestion.adm1_properties_vapor import (
    ADM1_vaporParameterBlock,
)
from watertap.property_models.anaerobic_digestion.modified_adm1_reactions import (
    ModifiedADM1ReactionParameterBlock,
    ModifiedADM1ReactionBlock,
)
from idaes.core.util.testing import initialization_tester

# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()


class TestParamBlock(object):
    @pytest.fixture(scope="class")
    def model(self):
        model = ConcreteModel()
        model.pparams = ModifiedADM1ParameterBlock()
        model.rparams = ModifiedADM1ReactionParameterBlock(
            property_package=model.pparams
        )

        return model

    @pytest.mark.unit
    def test_build(self, model):
        assert model.rparams.reaction_block_class is ModifiedADM1ReactionBlock

        assert len(model.rparams.rate_reaction_idx) == 25
        for i in model.rparams.rate_reaction_idx:
            assert i in [
                "R1",
                "R2",
                "R3",
                "R4",
                "R5",
                "R6",
                "R7",
                "R8",
                "R9",
                "R10",
                "R11",
                "R12",
                "R13",
                "R14",
                "R15",
                "R16",
                "R17",
                "R18",
                "R19",
                "R20",
                "R21",
                "R22",
                "R23",
                "R24",
                "R25",
            ]

        # Expected non-zero stoichiometries
        # Values from table 11 in reference

        mw_n = 14
        mw_c = 12
        mw_p = 31

        stoic = {
            # R1: Hydrolysis of carbohydrates
            ("R1", "Liq", "S_su"): 1,
            ("R1", "Liq", "X_ch"): -1,
            # R2:  Hydrolysis of proteins
            ("R2", "Liq", "S_aa"): 1,
            ("R2", "Liq", "X_pr"): -1,
            # R3:  Hydrolysis of lipids
            ("R3", "Liq", "S_su"): 0.05,
            ("R3", "Liq", "S_fa"): 0.95,
            ("R3", "Liq", "X_li"): -1,
            ("R3", "Liq", "S_IC"): 0.000029 * mw_c,
            ("R3", "Liq", "S_IP"): 0.0003441 * mw_p,
            # R4:  Uptake of sugars
            ("R4", "Liq", "S_su"): -1,
            ("R4", "Liq", "S_bu"): 0.1195,
            ("R4", "Liq", "S_pro"): 0.2422,
            ("R4", "Liq", "S_ac"): 0.3668,
            ("R4", "Liq", "S_h2"): 0.1715,
            ("R4", "Liq", "S_IC"): 0.00726 * mw_c,
            ("R4", "Liq", "S_IN"): -0.000615 * mw_n,
            ("R4", "Liq", "S_IP"): -0.000069 * mw_p,
            ("R4", "Liq", "X_su"): 0.10,
            # R5:  Uptake of amino acids
            ("R5", "Liq", "S_aa"): -1,
            ("R5", "Liq", "S_va"): 0.2116,
            ("R5", "Liq", "S_bu"): 0.2392,
            ("R5", "Liq", "S_pro"): 0.0460,
            ("R5", "Liq", "S_ac"): 0.3680,
            ("R5", "Liq", "S_h2"): 0.0552,
            ("R5", "Liq", "S_IC"): 0.00450 * mw_c,
            ("R5", "Liq", "S_IN"): 0.00741 * mw_n,
            ("R5", "Liq", "S_IP"): -0.0000556 * mw_p,
            ("R5", "Liq", "X_aa"): 0.08,
            # R6:  Uptake of long chain fatty acids (LCFAs)
            ("R6", "Liq", "S_fa"): -1,
            ("R6", "Liq", "S_ac"): 0.6580,
            ("R6", "Liq", "S_h2"): 0.2820,
            ("R6", "Liq", "S_IC"): -0.000989 * mw_c,
            ("R6", "Liq", "S_IN"): -0.000369 * mw_n,
            ("R6", "Liq", "S_IP"): -0.0000417 * mw_p,
            ("R6", "Liq", "X_fa"): 0.06,
            # R7:  Uptake of valerate
            ("R7", "Liq", "S_va"): -1,
            ("R7", "Liq", "S_pro"): 0.5076,
            ("R7", "Liq", "S_ac"): 0.2914,
            ("R7", "Liq", "S_h2"): 0.1410,
            ("R7", "Liq", "S_IC"): -0.000495 * mw_c,
            ("R7", "Liq", "S_IN"): -0.000369 * mw_n,
            ("R7", "Liq", "S_IP"): -0.0000417 * mw_p,
            ("R7", "Liq", "X_c4"): 0.06,
            # R8:  Uptake of butyrate
            ("R8", "Liq", "S_bu"): -1,
            ("R8", "Liq", "S_ac"): 0.7520,
            ("R8", "Liq", "S_h2"): 0.1880,
            ("R8", "Liq", "S_IC"): -0.000331 * mw_c,
            ("R8", "Liq", "S_IN"): -0.000369 * mw_n,
            ("R8", "Liq", "S_IP"): -0.0000417 * mw_p,
            ("R8", "Liq", "X_c4"): 0.06,
            # R9: Uptake of propionate
            ("R9", "Liq", "S_pro"): -1,
            ("R9", "Liq", "S_ac"): 0.5472,
            ("R9", "Liq", "S_h2"): 0.4128,
            ("R9", "Liq", "S_IC"): 0.008465 * mw_c,
            ("R9", "Liq", "S_IN"): -0.000246 * mw_n,
            ("R9", "Liq", "S_IP"): -0.0000278 * mw_p,
            ("R9", "Liq", "X_pro"): 0.04,
            # R10: Uptake of acetate
            ("R10", "Liq", "S_ac"): -1,
            ("R10", "Liq", "S_ch4"): 0.95,
            ("R10", "Liq", "S_IC"): 0.014881 * mw_c,
            ("R10", "Liq", "S_IN"): -0.000308 * mw_n,
            ("R10", "Liq", "S_IP"): -0.0000347 * mw_p,
            ("R10", "Liq", "X_ac"): 0.05,
            # R11: Uptake of hydrogen
            ("R11", "Liq", "S_h2"): -1,
            ("R11", "Liq", "S_ch4"): 0.94,
            ("R11", "Liq", "S_IC"): -0.016518 * mw_c,
            ("R11", "Liq", "S_IN"): -0.000369 * mw_n,
            ("R11", "Liq", "S_IP"): -0.0000417 * mw_p,
            ("R11", "Liq", "X_h2"): 0.06,
            # R12: Decay of X_su
            ("R12", "Liq", "S_IC"): 0.002773 * mw_c,
            ("R12", "Liq", "S_IN"): 0.003551 * mw_n,
            ("R12", "Liq", "S_IP"): 0.0005534 * mw_p,
            ("R12", "Liq", "X_ch"): 0.275,
            ("R12", "Liq", "X_pr"): 0.275,
            ("R12", "Liq", "X_li"): 0.35,
            ("R12", "Liq", "X_su"): -1,
            ("R12", "Liq", "X_I"): 0.1,
            # R13: Decay of X_aa
            ("R13", "Liq", "S_IC"): 0.002773 * mw_c,
            ("R13", "Liq", "S_IN"): 0.003551 * mw_n,
            ("R13", "Liq", "S_IP"): 0.0005534 * mw_p,
            ("R13", "Liq", "X_ch"): 0.275,
            ("R13", "Liq", "X_pr"): 0.275,
            ("R13", "Liq", "X_li"): 0.35,
            ("R13", "Liq", "X_aa"): -1,
            ("R13", "Liq", "X_I"): 0.1,
            # R14: Decay of X_fa
            ("R14", "Liq", "S_IC"): 0.002773 * mw_c,
            ("R14", "Liq", "S_IN"): 0.003551 * mw_n,
            ("R14", "Liq", "S_IP"): 0.0005534 * mw_p,
            ("R14", "Liq", "X_ch"): 0.275,
            ("R14", "Liq", "X_pr"): 0.275,
            ("R14", "Liq", "X_li"): 0.35,
            ("R14", "Liq", "X_fa"): -1,
            ("R14", "Liq", "X_I"): 0.1,
            # R15: Decay of X_c4
            ("R15", "Liq", "S_IC"): 0.002773 * mw_c,
            ("R15", "Liq", "S_IN"): 0.003551 * mw_n,
            ("R15", "Liq", "S_IP"): 0.0005534 * mw_p,
            ("R15", "Liq", "X_ch"): 0.275,
            ("R15", "Liq", "X_pr"): 0.275,
            ("R15", "Liq", "X_li"): 0.35,
            ("R15", "Liq", "X_c4"): -1,
            ("R15", "Liq", "X_I"): 0.1,
            # R16: Decay of X_pro
            ("R16", "Liq", "S_IC"): 0.002773 * mw_c,
            ("R16", "Liq", "S_IN"): 0.003551 * mw_n,
            ("R16", "Liq", "S_IP"): 0.0005534 * mw_p,
            ("R16", "Liq", "X_ch"): 0.275,
            ("R16", "Liq", "X_pr"): 0.275,
            ("R16", "Liq", "X_li"): 0.35,
            ("R16", "Liq", "X_pro"): -1,
            ("R16", "Liq", "X_I"): 0.1,
            # R17: Decay of X_ac
            ("R17", "Liq", "S_IC"): 0.002773 * mw_c,
            ("R17", "Liq", "S_IN"): 0.003551 * mw_n,
            ("R17", "Liq", "S_IP"): 0.0005534 * mw_p,
            ("R17", "Liq", "X_ch"): 0.275,
            ("R17", "Liq", "X_pr"): 0.275,
            ("R17", "Liq", "X_li"): 0.35,
            ("R17", "Liq", "X_ac"): -1,
            ("R17", "Liq", "X_I"): 0.1,
            # R18: Decay of X_h2
            ("R18", "Liq", "S_IC"): 0.002773 * mw_c,
            ("R18", "Liq", "S_IN"): 0.003551 * mw_n,
            ("R18", "Liq", "S_IP"): 0.0005534 * mw_p,
            ("R18", "Liq", "X_ch"): 0.275,
            ("R18", "Liq", "X_pr"): 0.275,
            ("R18", "Liq", "X_li"): 0.35,
            ("R18", "Liq", "X_h2"): -1,
            ("R18", "Liq", "X_I"): 0.1,
            # R19: Storage of S_va in X_PHA
            ("R19", "Liq", "S_va"): -1,
            ("R19", "Liq", "S_IC"): -0.000962 * mw_c,
            ("R19", "Liq", "S_IP"): 0.012903 * mw_p,
            ("R19", "Liq", "X_PHA"): 1,
            ("R19", "Liq", "X_PP"): -0.012903,
            ("R19", "Liq", "S_K"): 0.004301,
            ("R19", "Liq", "S_Mg"): 0.004301,
            # R20: Storage of S_bu in X_PHA
            ("R20", "Liq", "S_bu"): -1,
            ("R20", "Liq", "S_IP"): 0.012903 * mw_p,
            ("R20", "Liq", "X_PHA"): 1,
            ("R20", "Liq", "X_PP"): -0.012903,
            ("R20", "Liq", "S_K"): 0.004301,
            ("R20", "Liq", "S_Mg"): 0.004301,
            # R21: Storage of S_pro in X_PHA
            ("R21", "Liq", "S_pro"): -1,
            ("R21", "Liq", "S_IC"): 0.001786 * mw_c,
            ("R21", "Liq", "S_IP"): 0.012903 * mw_p,
            ("R21", "Liq", "X_PHA"): 1,
            ("R21", "Liq", "X_PP"): -0.012903,
            ("R21", "Liq", "S_K"): 0.004301,
            ("R21", "Liq", "S_Mg"): 0.004301,
            # R22: Storage of S_ac in X_PHA
            ("R22", "Liq", "S_ac"): -1,
            ("R22", "Liq", "S_IC"): 0.006250 * mw_c,
            ("R22", "Liq", "S_IP"): 0.012903 * mw_p,
            ("R22", "Liq", "X_PHA"): 1,
            ("R22", "Liq", "X_PP"): -0.012903,
            ("R22", "Liq", "S_K"): 0.004301,
            ("R22", "Liq", "S_Mg"): 0.004301,
            # R23: Lysis of X_PAO
            ("R23", "Liq", "S_IC"): 0.002773 * mw_c,
            ("R23", "Liq", "S_IN"): 0.003551 * mw_n,
            ("R23", "Liq", "S_IP"): 0.000553 * mw_p,
            ("R23", "Liq", "X_ch"): 0.275,
            ("R23", "Liq", "X_pr"): 0.275,
            ("R23", "Liq", "X_li"): 0.35,
            ("R23", "Liq", "X_I"): 0.1,
            ("R23", "Liq", "X_PAO"): -1,
            # R24: Lysis of X_PP
            ("R24", "Liq", "S_IP"): 1 * mw_p,
            ("R24", "Liq", "X_PP"): -1,
            ("R24", "Liq", "S_K"): 1 / 3,
            ("R24", "Liq", "S_Mg"): 1 / 3,
            # R25: Lysis of X_PAO
            ("R25", "Liq", "S_va"): 0.1,
            ("R25", "Liq", "S_bu"): 0.1,
            ("R25", "Liq", "S_pro"): 0.4,
            ("R25", "Liq", "S_ac"): 0.4,
            ("R25", "Liq", "S_IC"): -0.003118 * mw_c,
            ("R25", "Liq", "X_PHA"): -1,
        }

        assert len(model.rparams.rate_reaction_stoichiometry) == 25 * 32
        for i, v in model.rparams.rate_reaction_stoichiometry.items():

            assert i[0] in [
                "R1",
                "R2",
                "R3",
                "R4",
                "R5",
                "R6",
                "R7",
                "R8",
                "R9",
                "R10",
                "R11",
                "R12",
                "R13",
                "R14",
                "R15",
                "R16",
                "R17",
                "R18",
                "R19",
                "R20",
                "R21",
                "R22",
                "R23",
                "R24",
                "R25",
            ]
            if i in stoic:
                assert pytest.approx(stoic[i], rel=1e-2) == value(v)
            else:
                assert pytest.approx(value(v), rel=1e-2) == 0

        assert isinstance(model.rparams.Z_h2s, Param)
        assert isinstance(model.rparams.f_xi_xb, Var)
        assert value(model.rparams.f_xi_xb) == 0.1
        assert isinstance(model.rparams.f_ch_xb, Var)
        assert value(model.rparams.f_ch_xb) == 0.275
        assert isinstance(model.rparams.f_li_xb, Var)
        assert value(model.rparams.f_li_xb) == 0.35
        assert isinstance(model.rparams.f_pr_xb, Var)
        assert value(model.rparams.f_pr_xb) == 0.275
        assert isinstance(model.rparams.f_si_xb, Var)
        assert value(model.rparams.f_si_xb) == 0

        assert isinstance(model.rparams.f_fa_li, Var)
        assert value(model.rparams.f_fa_li) == 0.95
        assert isinstance(model.rparams.f_h2_su, Var)
        assert value(model.rparams.f_h2_su) == 0.1906
        assert isinstance(model.rparams.f_bu_su, Var)
        assert value(model.rparams.f_bu_su) == 0.1328
        assert isinstance(model.rparams.f_pro_su, Var)
        assert value(model.rparams.f_pro_su) == 0.2691
        assert isinstance(model.rparams.f_ac_su, Var)
        assert value(model.rparams.f_ac_su) == 0.4076
        assert isinstance(model.rparams.f_h2_aa, Var)
        assert value(model.rparams.f_h2_aa) == 0.06
        assert isinstance(model.rparams.f_va_aa, Var)
        assert value(model.rparams.f_va_aa) == 0.23
        assert isinstance(model.rparams.f_bu_aa, Var)
        assert value(model.rparams.f_bu_aa) == 0.26
        assert isinstance(model.rparams.f_pro_aa, Var)
        assert value(model.rparams.f_pro_aa) == 0.05
        assert isinstance(model.rparams.f_ac_aa, Var)
        assert value(model.rparams.f_ac_aa) == 0.40

        assert isinstance(model.rparams.Y_su, Var)
        assert value(model.rparams.Y_su) == 0.1
        assert isinstance(model.rparams.Y_aa, Var)
        assert value(model.rparams.Y_aa) == 0.08
        assert isinstance(model.rparams.Y_fa, Var)
        assert value(model.rparams.Y_fa) == 0.06
        assert isinstance(model.rparams.Y_c4, Var)
        assert value(model.rparams.Y_c4) == 0.06
        assert isinstance(model.rparams.Y_pro, Var)
        assert value(model.rparams.Y_pro) == 0.04
        assert isinstance(model.rparams.Y_ac, Var)
        assert value(model.rparams.Y_ac) == 0.05
        assert isinstance(model.rparams.Y_h2, Var)
        assert value(model.rparams.Y_h2) == 0.06

        assert isinstance(model.rparams.k_hyd_ch, Var)
        assert value(model.rparams.k_hyd_ch) == 10
        assert isinstance(model.rparams.k_hyd_pr, Var)
        assert value(model.rparams.k_hyd_pr) == 10
        assert isinstance(model.rparams.k_hyd_li, Var)
        assert value(model.rparams.k_hyd_li) == 10
        assert isinstance(model.rparams.K_S_IN, Var)
        assert value(model.rparams.K_S_IN) == 1e-4
        assert isinstance(model.rparams.k_m_su, Var)
        assert value(model.rparams.k_m_su) == 30
        assert isinstance(model.rparams.K_S_su, Var)
        assert value(model.rparams.K_S_su) == 0.5

        assert isinstance(model.rparams.pH_UL_aa, Var)
        assert value(model.rparams.pH_UL_aa) == 5.5
        assert isinstance(model.rparams.pH_LL_aa, Var)
        assert value(model.rparams.pH_LL_aa) == 4

        assert isinstance(model.rparams.k_m_aa, Var)
        assert value(model.rparams.k_m_aa) == 50
        assert isinstance(model.rparams.K_S_aa, Var)
        assert value(model.rparams.K_S_aa) == 0.3
        assert isinstance(model.rparams.k_m_fa, Var)
        assert value(model.rparams.k_m_fa) == 6
        assert isinstance(model.rparams.K_S_fa, Var)
        assert value(model.rparams.K_S_fa) == 0.4
        assert isinstance(model.rparams.K_I_h2_fa, Var)
        assert value(model.rparams.K_I_h2_fa) == 5e-6
        assert isinstance(model.rparams.k_m_c4, Var)
        assert value(model.rparams.k_m_c4) == 20
        assert isinstance(model.rparams.K_S_c4, Var)
        assert value(model.rparams.K_S_c4) == 0.2
        assert isinstance(model.rparams.K_I_h2_c4, Var)
        assert value(model.rparams.K_I_h2_c4) == 1e-5
        assert isinstance(model.rparams.k_m_pro, Var)
        assert value(model.rparams.k_m_pro) == 13
        assert isinstance(model.rparams.K_S_pro, Var)
        assert value(model.rparams.K_S_pro) == 0.1
        assert isinstance(model.rparams.K_I_h2_pro, Var)
        assert value(model.rparams.K_I_h2_pro) == 3.5e-6
        assert isinstance(model.rparams.k_m_ac, Var)
        assert value(model.rparams.k_m_ac) == 8
        assert isinstance(model.rparams.K_S_ac, Var)
        assert value(model.rparams.K_S_ac) == 0.15
        assert isinstance(model.rparams.K_I_nh3, Var)
        assert value(model.rparams.K_I_nh3) == 0.0018

        assert isinstance(model.rparams.pH_UL_ac, Var)
        assert value(model.rparams.pH_UL_ac) == 7
        assert isinstance(model.rparams.pH_LL_ac, Var)
        assert value(model.rparams.pH_LL_ac) == 6

        assert isinstance(model.rparams.k_m_h2, Var)
        assert value(model.rparams.k_m_h2) == 35
        assert isinstance(model.rparams.K_S_h2, Var)
        assert value(model.rparams.K_S_h2) == 7e-6

        assert isinstance(model.rparams.pH_UL_h2, Var)
        assert value(model.rparams.pH_UL_h2) == 6
        assert isinstance(model.rparams.pH_LL_h2, Var)
        assert value(model.rparams.pH_LL_h2) == 5

        assert isinstance(model.rparams.k_dec_X_su, Var)
        assert value(model.rparams.k_dec_X_su) == 0.02
        assert isinstance(model.rparams.k_dec_X_aa, Var)
        assert value(model.rparams.k_dec_X_aa) == 0.02
        assert isinstance(model.rparams.k_dec_X_fa, Var)
        assert value(model.rparams.k_dec_X_fa) == 0.02
        assert isinstance(model.rparams.k_dec_X_c4, Var)
        assert value(model.rparams.k_dec_X_c4) == 0.02
        assert isinstance(model.rparams.k_dec_X_pro, Var)
        assert value(model.rparams.k_dec_X_pro) == 0.02
        assert isinstance(model.rparams.k_dec_X_ac, Var)
        assert value(model.rparams.k_dec_X_ac) == 0.02
        assert isinstance(model.rparams.k_dec_X_h2, Var)
        assert value(model.rparams.k_dec_X_h2) == 0.02

        assert isinstance(model.rparams.K_a_va, Var)
        assert value(model.rparams.K_a_va) == 10 ** (-4.86)
        assert isinstance(model.rparams.K_a_bu, Var)
        assert value(model.rparams.K_a_bu) == 10 ** (-4.82)
        assert isinstance(model.rparams.K_a_pro, Var)
        assert value(model.rparams.K_a_pro) == 10 ** (-4.88)
        assert isinstance(model.rparams.K_a_ac, Var)
        assert value(model.rparams.K_a_ac) == 10 ** (-4.76)

        assert isinstance(model.rparams.K_I_h2s_ac, Var)
        assert value(model.rparams.K_I_h2s_ac) == 460e-3
        assert isinstance(model.rparams.K_I_h2s_c4, Var)
        assert value(model.rparams.K_I_h2s_c4) == 481e-3
        assert isinstance(model.rparams.K_I_h2s_h2, Var)
        assert value(model.rparams.K_I_h2s_h2) == 400e-3
        assert isinstance(model.rparams.K_I_h2s_pro, Var)
        assert value(model.rparams.K_I_h2s_pro) == 481e-3
        assert isinstance(model.rparams.K_S_IP, Var)
        assert value(model.rparams.K_S_IP) == 2e-5

        assert isinstance(model.rparams.b_PAO, Var)
        assert value(model.rparams.b_PAO) == 0.2
        assert isinstance(model.rparams.b_PHA, Var)
        assert value(model.rparams.b_PHA) == 0.2
        assert isinstance(model.rparams.b_PP, Var)
        assert value(model.rparams.b_PP) == 0.2

        assert isinstance(model.rparams.f_ac_PHA, Var)
        assert value(model.rparams.f_ac_PHA) == 0.4
        assert isinstance(model.rparams.f_bu_PHA, Var)
        assert value(model.rparams.f_bu_PHA) == 0.1
        assert isinstance(model.rparams.f_pro_PHA, Var)
        assert value(model.rparams.f_pro_PHA) == 0.4
        assert isinstance(model.rparams.f_va_PHA, Var)
        assert value(model.rparams.f_va_PHA) == 0.1

        assert isinstance(model.rparams.K_A, Var)
        assert value(model.rparams.K_A) == 4e-3
        assert isinstance(model.rparams.K_PP, Var)
        assert value(model.rparams.K_PP) == 0.32e-3
        assert isinstance(model.rparams.q_PHA, Var)
        assert value(model.rparams.q_PHA) == 3
        assert isinstance(model.rparams.Y_PO4, Var)
        assert value(model.rparams.Y_PO4) == 12.903e-3
        assert isinstance(model.rparams.K_XPP, Var)
        assert value(model.rparams.K_XPP) == 1 / 3
        assert isinstance(model.rparams.Mg_XPP, Var)
        assert value(model.rparams.Mg_XPP) == 1 / 3


class TestReactionBlock(object):
    @pytest.fixture(scope="class")
    def model(self):
        model = ConcreteModel()
        model.pparams = ModifiedADM1ParameterBlock()
        model.vparams = ADM1_vaporParameterBlock()
        model.rparams = ModifiedADM1ReactionParameterBlock(
            property_package=model.pparams
        )

        model.props = model.pparams.build_state_block([1])
        model.props_vap = model.vparams.build_state_block([1])
        model.rxns = model.rparams.build_reaction_block([1], state_block=model.props)

        return model

    @pytest.mark.unit
    def test_build(self, model):
        assert model.rxns[1].conc_mass_comp_ref is model.props[1].conc_mass_comp

    @pytest.mark.unit
    def test_rxn_rate(self, model):
        assert isinstance(model.rxns[1].reaction_rate, Var)
        assert len(model.rxns[1].reaction_rate) == 25
        assert isinstance(model.rxns[1].rate_expression, Constraint)
        assert len(model.rxns[1].rate_expression) == 25

    @pytest.mark.unit
    def test_get_reaction_rate_basis(self, model):
        assert model.rxns[1].get_reaction_rate_basis() == MaterialFlowBasis.mass

    @pytest.mark.component
    def test_initialize(self, model):
        assert model.rxns.initialize() is None

    @pytest.mark.unit
    def check_units(self, model):
        assert_units_consistent(model)


class TestReactor:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()

        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.props = ModifiedADM1ParameterBlock()
        m.fs.props_vap = ADM1_vaporParameterBlock()
        m.fs.rxn_props = ModifiedADM1ReactionParameterBlock(property_package=m.fs.props)

        m.fs.unit = AD(
            liquid_property_package=m.fs.props,
            vapor_property_package=m.fs.props_vap,
            reaction_package=m.fs.rxn_props,
            has_heat_transfer=True,
            has_pressure_change=False,
        )

        # Feed conditions based on mass balance in Flores-Alsina, where 0 terms are expressed as 1e-9
        m.fs.unit.inlet.flow_vol.fix(0.001967593)  # Double check this value
        m.fs.unit.inlet.temperature.fix(308.15)
        m.fs.unit.inlet.pressure.fix(101325)

        m.fs.unit.inlet.conc_mass_comp[0, "S_su"].fix(0.034597)
        m.fs.unit.inlet.conc_mass_comp[0, "S_aa"].fix(0.015037)
        m.fs.unit.inlet.conc_mass_comp[0, "S_fa"].fix(1e-8)
        m.fs.unit.inlet.conc_mass_comp[0, "S_va"].fix(1e-8)
        m.fs.unit.inlet.conc_mass_comp[0, "S_bu"].fix(1e-8)
        m.fs.unit.inlet.conc_mass_comp[0, "S_pro"].fix(1e-8)
        m.fs.unit.inlet.conc_mass_comp[0, "S_ac"].fix(0.025072)
        m.fs.unit.inlet.conc_mass_comp[0, "S_h2"].fix(1e-8)
        m.fs.unit.inlet.conc_mass_comp[0, "S_ch4"].fix(1e-8)
        m.fs.unit.inlet.conc_mass_comp[0, "S_IC"].fix(0.34628)
        m.fs.unit.inlet.conc_mass_comp[0, "S_IN"].fix(0.60014)
        m.fs.unit.inlet.conc_mass_comp[0, "S_IP"].fix(0.22677)
        m.fs.unit.inlet.conc_mass_comp[0, "S_I"].fix(0.026599)

        m.fs.unit.inlet.conc_mass_comp[0, "X_ch"].fix(7.3687)
        m.fs.unit.inlet.conc_mass_comp[0, "X_pr"].fix(7.7308)
        m.fs.unit.inlet.conc_mass_comp[0, "X_li"].fix(10.3288)
        m.fs.unit.inlet.conc_mass_comp[0, "X_su"].fix(1e-8)
        m.fs.unit.inlet.conc_mass_comp[0, "X_aa"].fix(1e-8)
        m.fs.unit.inlet.conc_mass_comp[0, "X_fa"].fix(1e-8)
        m.fs.unit.inlet.conc_mass_comp[0, "X_c4"].fix(1e-8)
        m.fs.unit.inlet.conc_mass_comp[0, "X_pro"].fix(1e-8)
        m.fs.unit.inlet.conc_mass_comp[0, "X_ac"].fix(1e-8)
        m.fs.unit.inlet.conc_mass_comp[0, "X_h2"].fix(1e-8)
        m.fs.unit.inlet.conc_mass_comp[0, "X_I"].fix(12.7727)
        m.fs.unit.inlet.conc_mass_comp[0, "X_PHA"].fix(0.0022493)
        m.fs.unit.inlet.conc_mass_comp[0, "X_PP"].fix(1.04110)
        m.fs.unit.inlet.conc_mass_comp[0, "X_PAO"].fix(3.4655)
        m.fs.unit.inlet.conc_mass_comp[0, "S_K"].fix(0.02268)
        m.fs.unit.inlet.conc_mass_comp[0, "S_Mg"].fix(0.02893)

        m.fs.unit.inlet.cations[0].fix(0.04)
        m.fs.unit.inlet.anions[0].fix(0.02)

        m.fs.unit.volume_liquid.fix(3400)
        m.fs.unit.volume_vapor.fix(300)
        m.fs.unit.liquid_outlet.temperature.fix(308.15)

        # Touch on-demand property, TSS, at inlet and outlet
        m.fs.unit.liquid_phase.properties_in[0].TSS
        m.fs.unit.liquid_phase.properties_out[0].TSS

        return m

    @pytest.mark.unit
    def test_dof(self, model):
        assert degrees_of_freedom(model) == 0

    @pytest.mark.unit
    def test_unit_consistency(self, model):
        assert_units_consistent(model)

    @pytest.mark.unit
    def test_scaling_factors(self, model):
        m = model
        iscale.calculate_scaling_factors(m)

        # check that all variables have scaling factors
        unscaled_var_list = list(iscale.unscaled_variables_generator(m))
        assert len(unscaled_var_list) == 0

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    @pytest.mark.requires_idaes_solver
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    @pytest.mark.requires_idaes_solver
    def test_solve(self, model):
        solver = get_solver()
        results = solver.solve(model)

        assert_optimal_termination(results)

    # TO DO: retest after conversion changes
    @pytest.mark.component
    @pytest.mark.requires_idaes_solver
    def test_solution(self, model):
        assert value(model.fs.unit.liquid_outlet.flow_vol[0]) == pytest.approx(
            0.0019675, rel=1e-2
        )
        assert value(model.fs.unit.liquid_outlet.temperature[0]) == pytest.approx(
            308.15, rel=1e-2
        )
        assert value(model.fs.unit.liquid_outlet.pressure[0]) == pytest.approx(
            101325, rel=1e-2
        )
        assert value(
            model.fs.unit.liquid_outlet.conc_mass_comp[0, "S_su"]
        ) == pytest.approx(8.744124, rel=1e-2)
        assert value(
            model.fs.unit.liquid_outlet.conc_mass_comp[0, "S_aa"]
        ) == pytest.approx(5.31e-3, rel=1e-2)
        assert value(
            model.fs.unit.liquid_outlet.conc_mass_comp[0, "S_fa"]
        ) == pytest.approx(10.7453, rel=1e-2)
        assert value(
            model.fs.unit.liquid_outlet.conc_mass_comp[0, "S_va"]
        ) == pytest.approx(0.015746, rel=1e-2)
        assert value(
            model.fs.unit.liquid_outlet.conc_mass_comp[0, "S_bu"]
        ) == pytest.approx(0.0163891, rel=1e-2)
        assert value(
            model.fs.unit.liquid_outlet.conc_mass_comp[0, "S_pro"]
        ) == pytest.approx(0.0356794, rel=1e-2)
        assert value(
            model.fs.unit.liquid_outlet.conc_mass_comp[0, "S_ac"]
        ) == pytest.approx(0.0431179, rel=1e-2)
        assert value(
            model.fs.unit.liquid_outlet.conc_mass_comp[0, "S_h2"]
        ) == pytest.approx(0.0112939, rel=1e-2)
        assert value(
            model.fs.unit.liquid_outlet.conc_mass_comp[0, "S_ch4"]
        ) == pytest.approx(3.28736e-7, abs=1e-5)
        assert value(
            model.fs.unit.liquid_outlet.conc_mass_comp[0, "S_IC"]
        ) == pytest.approx(0.09373 * 12, rel=1e-2)
        assert value(
            model.fs.unit.liquid_outlet.conc_mass_comp[0, "S_IN"]
        ) == pytest.approx(0.11650 * 14, rel=1e-2)
        assert value(
            model.fs.unit.liquid_outlet.conc_mass_comp[0, "S_IP"]
        ) == pytest.approx(29.11478, rel=1e-2)
        assert value(
            model.fs.unit.liquid_outlet.conc_mass_comp[0, "S_I"]
        ) == pytest.approx(0.02660, rel=1e-2)
        assert value(
            model.fs.unit.liquid_outlet.conc_mass_comp[0, "X_ch"]
        ) == pytest.approx(0.0407196, rel=1e-2)
        assert value(
            model.fs.unit.liquid_outlet.conc_mass_comp[0, "X_pr"]
        ) == pytest.approx(0.0424521, rel=1e-2)
        assert value(
            model.fs.unit.liquid_outlet.conc_mass_comp[0, "X_li"]
        ) == pytest.approx(0.0565536, rel=1e-2)
        assert value(
            model.fs.unit.liquid_outlet.conc_mass_comp[0, "X_su"]
        ) == pytest.approx(8.369087e-10, abs=1e-5)
        assert value(
            model.fs.unit.liquid_outlet.conc_mass_comp[0, "X_aa"]
        ) == pytest.approx(0.486516, rel=1e-2)
        assert value(
            model.fs.unit.liquid_outlet.conc_mass_comp[0, "X_fa"]
        ) == pytest.approx(4.35411e-6, abs=1e-5)
        assert value(
            model.fs.unit.liquid_outlet.conc_mass_comp[0, "X_c4"]
        ) == pytest.approx(4.0393379e-6, abs=1e-5)
        assert value(
            model.fs.unit.liquid_outlet.conc_mass_comp[0, "X_pro"]
        ) == pytest.approx(4.18807e-6, abs=1e-5)
        assert value(
            model.fs.unit.liquid_outlet.conc_mass_comp[0, "X_ac"]
        ) == pytest.approx(4.274973e-6, abs=1e-5)
        assert value(
            model.fs.unit.liquid_outlet.conc_mass_comp[0, "X_h2"]
        ) == pytest.approx(1.378092e-8, abs=1e-5)
        assert value(
            model.fs.unit.liquid_outlet.conc_mass_comp[0, "X_I"]
        ) == pytest.approx(13.0695, rel=1e-2)
        assert value(
            model.fs.unit.liquid_outlet.conc_mass_comp[0, "X_PHA"]
        ) == pytest.approx(7.2792898, rel=1e-2)
        assert value(
            model.fs.unit.liquid_outlet.conc_mass_comp[0, "X_PP"]
        ) == pytest.approx(0.1143025, rel=1e-2)
        assert value(
            model.fs.unit.liquid_outlet.conc_mass_comp[0, "X_PAO"]
        ) == pytest.approx(0.69310, rel=1e-2)
        assert value(
            model.fs.unit.liquid_outlet.conc_mass_comp[0, "S_K"]
        ) == pytest.approx(0.33162, rel=1e-2)
        assert value(
            model.fs.unit.liquid_outlet.conc_mass_comp[0, "S_Mg"]
        ) == pytest.approx(0.33787, rel=1e-2)
        assert value(model.fs.unit.liquid_outlet.anions[0]) == pytest.approx(
            2e-2, rel=1e-2
        )
        assert value(model.fs.unit.liquid_outlet.cations[0]) == pytest.approx(
            4e-2, rel=1e-2
        )

        assert value(
            model.fs.unit.liquid_phase.reactions[0].conc_mass_va
        ) == pytest.approx(0.0157436, rel=1e-2)
        assert value(
            model.fs.unit.liquid_phase.reactions[0].conc_mass_bu
        ) == pytest.approx(0.0163960, rel=1e-2)
        assert value(
            model.fs.unit.liquid_phase.reactions[0].conc_mass_ac
        ) == pytest.approx(0.0431131, rel=1e-2)
        assert value(
            model.fs.unit.liquid_phase.reactions[0].conc_mass_pro
        ) == pytest.approx(0.03567417, rel=1e-2)
        assert value(
            model.fs.unit.liquid_phase.reactions[0].conc_mol_hco3
        ) == pytest.approx(0.09336803, rel=1e-2)
        assert value(
            model.fs.unit.liquid_phase.reactions[0].conc_mol_nh3
        ) == pytest.approx(0.04271418, rel=1e-2)
        assert value(
            model.fs.unit.liquid_phase.reactions[0].conc_mol_co2
        ) == pytest.approx(0.00036273, rel=1e-2)
        assert value(
            model.fs.unit.liquid_phase.reactions[0].conc_mol_nh4
        ) == pytest.approx(0.07378798, rel=1e-2)
        assert value(
            model.fs.unit.liquid_phase.reactions[0].conc_mol_Mg
        ) == pytest.approx(0.00038049, rel=1e-2)
        assert value(
            model.fs.unit.liquid_phase.reactions[0].conc_mol_K
        ) == pytest.approx(0.00038049, rel=1e-2)

        assert value(model.fs.unit.liquid_phase.reactions[0].S_H) == pytest.approx(
            1.9180052e-9, rel=1e-2
        )
        assert value(model.fs.unit.liquid_phase.reactions[0].pKW, Var) == pytest.approx(
            -log10(2.08e-14), rel=1e-2
        )
        assert value(
            model.fs.unit.liquid_phase.reactions[0].pK_a_co2, Var
        ) == pytest.approx(-log10(4.94e-7), rel=1e-2)
        assert value(
            model.fs.unit.liquid_phase.reactions[0].pK_a_IN, Var
        ) == pytest.approx(-log10(1.11e-9), rel=1e-2)
        assert value(
            1
            - model.fs.unit.liquid_phase.properties_out[0].TSS
            / model.fs.unit.liquid_phase.properties_in[0].TSS
        ) * 100 == pytest.approx(55.989195, rel=1e-4)
        assert value(
            1
            - model.fs.unit.liquid_phase.properties_out[0].VSS
            / model.fs.unit.liquid_phase.properties_in[0].VSS
        ) * 100 == pytest.approx(51.72925597, rel=1e-4)
        assert value(
            1
            - model.fs.unit.liquid_phase.properties_out[0].ISS
            / model.fs.unit.liquid_phase.properties_in[0].ISS
        ) * 100 == pytest.approx(86.582365, rel=1e-4)



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Tests for zero-order evaporation pond model
"""
import pytest


from pyomo.environ import (
    Block,
    check_optimal_termination,
    ConcreteModel,
    Constraint,
    value,
    Var,
)
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock
from idaes.core.solvers import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.testing import initialization_tester
from idaes.core import UnitModelCostingBlock

from watertap.unit_models.zero_order import EvaporationPondZO
from watertap.core.wt_database import Database
from watertap.core.zero_order_properties import WaterParameterBlock
from watertap.costing.zero_order_costing import ZeroOrderCosting

solver = get_solver()


class TestEvaporationPondZO:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.db = Database()

        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.params = WaterParameterBlock(
            solute_list=["tds", "magnesium", "calcium", "nitrate", "sulfate", "tss"]
        )

        m.fs.unit = EvaporationPondZO(property_package=m.fs.params, database=m.db)

        m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(10)
        m.fs.unit.inlet.flow_mass_comp[0, "tds"].fix(123)
        m.fs.unit.inlet.flow_mass_comp[0, "magnesium"].fix(456)
        m.fs.unit.inlet.flow_mass_comp[0, "calcium"].fix(789)
        m.fs.unit.inlet.flow_mass_comp[0, "nitrate"].fix(10)
        m.fs.unit.inlet.flow_mass_comp[0, "sulfate"].fix(11)
        m.fs.unit.inlet.flow_mass_comp[0, "tss"].fix(12)

        return m

    @pytest.mark.unit
    def test_build(self, model):
        assert model.fs.unit.config.database is model.db
        assert model.fs.unit._tech_type == "evaporation_pond"

        assert isinstance(model.fs.unit.air_temperature, Var)
        assert isinstance(model.fs.unit.solar_radiation, Var)
        assert isinstance(model.fs.unit.evaporation_rate_adj_factor, Var)
        assert isinstance(model.fs.unit.evap_rate_calc_a_parameter, Var)
        assert isinstance(model.fs.unit.evap_rate_calc_b_parameter, Var)
        assert isinstance(model.fs.unit.evap_rate_calc_c_parameter, Var)
        assert isinstance(model.fs.unit.adj_area_calc_a_parameter, Var)
        assert isinstance(model.fs.unit.adj_area_calc_b_parameter, Var)
        assert isinstance(model.fs.unit.area, Var)
        assert isinstance(model.fs.unit.adj_area, Var)
        assert isinstance(model.fs.unit.evaporation_rate_pure, Var)
        assert isinstance(model.fs.unit.evaporation_rate_salt, Var)
        assert isinstance(model.fs.unit.evap_rate_pure_constraint, Constraint)
        assert isinstance(model.fs.unit.evap_rate_salt_constraint, Constraint)
        assert isinstance(model.fs.unit.area_constraint, Constraint)
        assert isinstance(model.fs.unit.area_adj_constraint, Constraint)

    @pytest.mark.component
    def test_load_parameters(self, model):
        data = model.db.get_unit_operation_parameters("evaporation_pond")
        model.fs.unit.load_parameters_from_database(use_default_removal=True)
        assert model.fs.unit.recovery_frac_mass_H2O[0].fixed
        assert model.fs.unit.recovery_frac_mass_H2O[0].value == 0.0001

        for (t, j), v in model.fs.unit.removal_frac_mass_comp.items():
            assert v.fixed
            if j not in data["removal_frac_mass_comp"]:
                assert v.value == data["default_removal_frac_mass_comp"]["value"]
            else:
                assert v.value == data["removal_frac_mass_comp"][j]["value"]

        assert model.fs.unit.air_temperature[0].fixed
        assert (
            model.fs.unit.air_temperature[0].value == data["air_temperature"]["value"]
        )
        assert model.fs.unit.solar_radiation[0].fixed
        assert (
            model.fs.unit.solar_radiation[0].value == data["solar_radiation"]["value"]
        )
        assert model.fs.unit.dike_height[0].fixed
        assert model.fs.unit.dike_height[0].value == data["dike_height"]["value"]
        assert model.fs.unit.evaporation_rate_adj_factor[0].fixed
        assert (
            model.fs.unit.evaporation_rate_adj_factor[0].value
            == data["evaporation_rate_adj_factor"]["value"]
        )
        assert model.fs.unit.evap_rate_calc_a_parameter[0].fixed
        assert (
            model.fs.unit.evap_rate_calc_a_parameter[0].value
            == data["evap_rate_calc_a_parameter"]["value"]
        )
        assert model.fs.unit.evap_rate_calc_b_parameter[0].fixed
        assert (
            model.fs.unit.evap_rate_calc_b_parameter[0].value
            == data["evap_rate_calc_b_parameter"]["value"]
        )
        assert model.fs.unit.evap_rate_calc_c_parameter[0].fixed
        assert (
            model.fs.unit.evap_rate_calc_c_parameter[0].value
            == data["evap_rate_calc_c_parameter"]["value"]
        )
        assert model.fs.unit.adj_area_calc_a_parameter[0].fixed
        assert (
            model.fs.unit.adj_area_calc_a_parameter[0].value
            == data["adj_area_calc_a_parameter"]["value"]
        )
        assert model.fs.unit.adj_area_calc_b_parameter[0].fixed
        assert (
            model.fs.unit.adj_area_calc_b_parameter[0].value
            == data["adj_area_calc_b_parameter"]["value"]
        )

        assert model.fs.unit.energy_electric_flow_vol_inlet.fixed
        assert (
            model.fs.unit.energy_electric_flow_vol_inlet.value
            == data["energy_electric_flow_vol_inlet"]["value"]
        )

    @pytest.mark.component
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model.fs.unit) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model.fs.unit)

    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        assert pytest.approx(1.411, rel=1e-5) == value(
            model.fs.unit.properties_in[0].flow_vol
        )
        assert pytest.approx(7.087172, rel=1e-5) == value(
            model.fs.unit.properties_in[0].conc_mass_comp["H2O"]
        )
        assert pytest.approx(87.17221, rel=1e-5) == value(
            model.fs.unit.properties_in[0].conc_mass_comp["tds"]
        )
        assert pytest.approx(323.175053, rel=1e-5) == value(
            model.fs.unit.properties_in[0].conc_mass_comp["magnesium"]
        )
        assert pytest.approx(559.177888, rel=1e-5) == value(
            model.fs.unit.properties_in[0].conc_mass_comp["calcium"]
        )
        assert pytest.approx(7.087172, rel=1e-5) == value(
            model.fs.unit.properties_in[0].conc_mass_comp["nitrate"]
        )
        assert pytest.approx(7.795889, rel=1e-5) == value(
            model.fs.unit.properties_in[0].conc_mass_comp["sulfate"]
        )
        assert pytest.approx(8.504606, rel=1e-5) == value(
            model.fs.unit.properties_in[0].conc_mass_comp["tss"]
        )

        assert pytest.approx(0.028021, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].flow_vol
        )
        assert pytest.approx(0.0356878, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["H2O"]
        )
        assert pytest.approx(87.791299, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["tds"]
        )
        assert pytest.approx(325.470182, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["magnesium"]
        )
        assert pytest.approx(563.149066, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["calcium"]
        )
        assert pytest.approx(7.13750, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["nitrate"]
        )
        assert pytest.approx(7.8512544, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["sulfate"]
        )
        assert pytest.approx(8.5650048, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["tss"]
        )

        assert pytest.approx(1.382979, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].flow_vol
        )
        assert pytest.approx(7.2300447, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].conc_mass_comp["H2O"]
        )
        assert pytest.approx(87.15967, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].conc_mass_comp["tds"]
        )
        assert pytest.approx(323.128550, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].conc_mass_comp["magnesium"]
        )
        assert pytest.approx(559.097426, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].conc_mass_comp["calcium"]
        )
        assert pytest.approx(7.086152, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].conc_mass_comp["nitrate"]
        )
        assert pytest.approx(7.794767, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].conc_mass_comp["sulfate"]
        )
        assert pytest.approx(8.503382, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].conc_mass_comp["tss"]
        )

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, model):
        for j in model.fs.params.component_list:
            assert 1e-6 >= abs(
                value(
                    model.fs.unit.inlet.flow_mass_comp[0, j]
                    - model.fs.unit.treated.flow_mass_comp[0, j]
                    - model.fs.unit.byproduct.flow_mass_comp[0, j]
                )
            )

    @pytest.mark.component
    def test_report(self, model):

        model.fs.unit.report()


def test_costing():
    m = ConcreteModel()
    m.db = Database()

    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.params = WaterParameterBlock(
        solute_list=["tds", "magnesium", "calcium", "nitrate", "sulfate", "tss"]
    )
    m.fs.costing = ZeroOrderCosting()
    m.fs.unit = EvaporationPondZO(property_package=m.fs.params, database=m.db)

    m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(10)
    m.fs.unit.inlet.flow_mass_comp[0, "tds"].fix(123)
    m.fs.unit.inlet.flow_mass_comp[0, "magnesium"].fix(456)
    m.fs.unit.inlet.flow_mass_comp[0, "calcium"].fix(789)
    m.fs.unit.inlet.flow_mass_comp[0, "nitrate"].fix(10)
    m.fs.unit.inlet.flow_mass_comp[0, "sulfate"].fix(11)
    m.fs.unit.inlet.flow_mass_comp[0, "tss"].fix(12)
    m.fs.unit.load_parameters_from_database(use_default_removal=True)

    m.fs.unit.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)

    assert isinstance(m.fs.costing.evaporation_pond, Block)
    assert isinstance(m.fs.costing.evaporation_pond.cost_per_acre_a_parameter, Var)
    assert isinstance(m.fs.costing.evaporation_pond.cost_per_acre_b_parameter, Var)
    assert isinstance(m.fs.costing.evaporation_pond.cost_per_acre_c_parameter, Var)
    assert isinstance(m.fs.costing.evaporation_pond.cost_per_acre_d_parameter, Var)
    assert isinstance(m.fs.costing.evaporation_pond.cost_per_acre_e_parameter, Var)
    assert isinstance(m.fs.costing.evaporation_pond.liner_thickness, Var)
    assert isinstance(m.fs.costing.evaporation_pond.land_cost, Var)
    assert isinstance(m.fs.costing.evaporation_pond.land_clearing_cost, Var)

    assert isinstance(m.fs.unit.costing.capital_cost, Var)
    assert isinstance(m.fs.unit.costing.capital_cost_constraint, Constraint)

    assert_units_consistent(m.fs)
    assert degrees_of_freedom(m.fs.unit) == 0
